import inspect, io
from qtpy.QtWidgets import QWidget, QVBoxLayout
from qtpy.QtCore import QObject, Signal
from superqt import QCollapsible
from magicgui.widgets import Container, ComboBox, ListEdit, CheckBox, LineEdit, FileEdit, PushButton, create_widget
from magicgui.type_map import get_widget_class
from contextlib import redirect_stdout, redirect_stderr
from functools import partial
from pathlib import Path
from ..crappifiers import AdditiveGaussian, Poisson, SaltPepper
from ..models import ResUNet, RDResUNet, SwinIR
from ..data import ImageDataset, SlidingDataset, PairedImageDataset, PairedSlidingDataset

ADVANCED_PARAMS = {
    ResUNet : list(range(4,6+1)),
    RDResUNet : list(range(4,6+1)) + list(range(12,16+1)),
    SwinIR : list(range(8,19+1)),
}

IGNORE_PARAMS = {
    SwinIR : [15],
    ImageDataset : [3,9],
    SlidingDataset : [3,12],
    PairedImageDataset : [9],
    PairedSlidingDataset : [12],
}

class ObjectEdit(QWidget):
    def __init__(self, title : str, objects : list, hide_crappifier : bool = False):
        super().__init__()
        
        self.collapse = QCollapsible(title)

        self.type = ComboBox(name="Type", choices=[item.__name__ for item in objects])
        self.type.changed.connect(self._clear_arguments)
        self.collapse.addWidget(self.type.native)

        self.arg_container = Container()
        self.collapse.addWidget(self.arg_container.native)

        # If object is dataset picker
        if any(item in objects for item in [ImageDataset, SlidingDataset]):
            self.crappifier = ObjectEdit("Crappifier", [AdditiveGaussian, Poisson, SaltPepper])
            self.crappifier.type.changed.connect(lambda : self._set_arguments("crappifier", self.crappifier.object, raw=True))
            self.crappifier.arg_container.changed.connect(lambda : self._set_arguments("crappifier", self.crappifier.object, raw=True))

            self.type.changed.connect(self._assert_crappifier)
            self.collapse.addWidget(self.crappifier)
            if hide_crappifier:
                self.crappifier.collapse.collapse()

        self.advanced_container = Container()
        self.advanced_collapse = QCollapsible("Advanced Options")
        self.advanced_collapse.addWidget(self.advanced_container.native)
        self.collapse.addWidget(self.advanced_collapse)

        # self.debug = PushButton(text="DEBUG: Get Object")
        # self.debug.changed.connect(lambda x : print(self.object))
        # self.collapse.addWidget(self.debug.native)

        self.collapse.expand()
        self._clear_arguments()

        layout = QVBoxLayout()
        layout.addWidget(self.collapse)
        self.setLayout(layout)
    
    @property
    def object(self):
        return f"{self.type.current_choice}({', '.join([f'{key}={value}' for key, value in self.arguments.items()])})"

    def _clear_arguments(self):
        self.arguments = {}
        self.arg_container.clear()
        self.advanced_container.clear()

        choice = eval(self.type.current_choice)
        spec = inspect.getfullargspec(choice)
        advanced_idx = ADVANCED_PARAMS.get(choice, [])
        ignore_idx = IGNORE_PARAMS.get(choice, [])

        # Parameters without a default are None
        defaults = [None]*(len(spec.args[1:])-len(spec.defaults)) + list(spec.defaults)

        for idx, (arg, default) in enumerate(zip(spec.args[1:], defaults)):
            if idx in ignore_idx:
                continue

            widget_type = get_widget_class(annotation=spec.annotations[arg])[0]

            if widget_type is not ListEdit:
                if widget_type not in [CheckBox, LineEdit, FileEdit]:
                    options = dict(max=2**14)
                if widget_type is FileEdit:
                    options = dict(mode="d")
                else:
                    options = {}
                widget = create_widget(value=default, annotation=spec.annotations[arg], name=arg, options=options)
            elif str(spec.annotations[arg]).count("list") == 1:
                if type(default) is not list and default:
                    default = [default]
                widget = ListEdit(value=default if default is not None else [0], name=arg, options=(dict(max=2**14, min=min(default[0] if default else 0, 0)) if any(item in str(spec.annotations[arg]) for item in ["int", "float"]) else {}))
                if default is None:
                    widget._pop_value()
            else:
                if type(default) is not list and default:
                    default = [default]
                widget = _LargeList(value=default if default is not None else [[0]], name=arg, options=dict(max=2**14))
                widget._pop_value()

            widget.changed.connect(partial(self._set_arguments, arg))
            self._set_arguments(arg, default)

            if idx in advanced_idx:
                self.advanced_container.append(widget)
            else:
                self.arg_container.append(widget)
        
        # Crappifier widget already exists, only need to get current state
        if choice in [ImageDataset, SlidingDataset]:
            self._set_arguments("crappifier", self.crappifier.object, raw=True)

        if len(advanced_idx) > 0:
            self.advanced_collapse.show()
        else:
            self.advanced_collapse.hide()
    
    def _set_arguments(self, name, value, raw=False):
        self.arguments[name] = (value if not issubclass(type(value), (str, Path)) or raw else f'"{value}"') if value != [] else None

    def _assert_crappifier(self):
        if eval(self.type.current_choice) in [ImageDataset, SlidingDataset]:
            self.crappifier.show()
        else:
            self.crappifier.hide()

class _LargeList(ListEdit):
    def __init__(self, value, name : str, options : dict = None):
        super().__init__(name=name, layout="vertical")
        self.list_options = {} if options is None else options

        head = self._make_list(value)
        for item in head:
            idx = (len(self)-2)
            self.insert(idx, item)

    def _append_value(self, widget = None):
        if widget == None:
            widget=ListEdit(value=[0], options=self.list_options)
        idx = len(self)-2
        self.insert(idx, widget)
        self[idx].changed.connect(lambda : self.changed.emit(self.value))

        self.changed.emit(self.value)
    
    def _make_list(self, value):
        if type(value[0]) is list:
            return [self._make_list(item) for item in value]
        
        return ListEdit(value=value, options=self.list_options)
    
class SignalWrapper(QObject):

    out = Signal(str)
    err = Signal(str)

    def __init__(self, func):
        super().__init__()
        self.func = func

    def capture(self):
        with redirect_stdout(_SignalCapture(self.out)):
            with redirect_stderr(_SignalCapture(self.err)):
                self.func()
    
class _SignalCapture():
    def __init__(self, signal):
        self.signal = signal

    def write(self, text):
        # Called on every sys.stdout
        self.signal.emit(text)



import torch, glob, os, random, czifile, tifffile, psutil, warnings
import numpy as np
from torch.utils.data import Dataset
from pathlib import Path
from tqdm import tqdm
from PIL import Image
from .crappifiers import Crappifier, Poisson
from .util import _force_list

class ImageDataset(Dataset):
    def __init__(self, path : Path, hr_res : int = 512, lr_scale : int = 4, crappifier : Crappifier = Poisson(), n_frames : list[int] = -1, extension : str = "tif", val_split : float = 0.1, rotation : bool = True, split_seed : int = 0, transforms : list[torch.nn.Module] = None):
        r"""Training dataset for loading high-resolution images from individual files and returning high-low-resolution pairs, the latter receiving crappification.

        Dataset used for pre-tiled image files. For image sheets (e.g. .czi files), use :class:`SlidingDataset`.

        LR mode (dataset loads only unmodified low-resolution images for prediction) can be enabled by
        either inputting images less than or equal to LR size (``hr_res``/``lr_scale``) or by setting ``lr_scale`` = None and ``hr_res`` = LR resolution.

        Args:
            path (Path) : Path to folder containing high-resolution images. Can also be a str.

            hr_res (int) : Resolution of high-resolution images. Images larger than this will be downscaled to this resolution. Images smaller will be padded. Default is 512.

            lr_scale (int) : Downscaling factor for low-resolution images to simulate undersampling. Choose a power of 2 for best results. Default is 4.

            crappifier (Crappifier) : Crappifier for degrading low-resolution images to simulate undersampling. Not used in LR mode. Default is :class:`Poisson`.

            n_frames (list[int]) : Amount of stacked frames per image, disregarding color channels. Can also be list of low-resolution and high-resolution stack amounts respectively. A value of -1 uses all stacked image frames. Default is -1.

            extension (str) : File extension of images. Default is "tif".

            val_split (float) : Proportion of images to be held out for evaluation/prediction. Default is 0.1.

            rotation (bool) : Whether to randomly rotate and/or flip images when loading data. Only used during training if applicable. Default is True.

            split_seed (int) : Seed for random train/evaluation data splitting. A value of None splits the last images as evaluation. Default is 0.

            transforms (list[nn.Module]) : Additional final data transforms to apply. Default is None.
        """
        super().__init__()
        self.path = Path(path) if type(path) is str else path
        if not path or not self.path.exists(): raise FileNotFoundError(f'Path "{self.path}" does not exist.')

        self.hr_files = _root_glob(f"*.{extension}", root_dir=self.path)
        if not len(self.hr_files) > 0: raise FileNotFoundError(f'No .{extension} files exist in path "{self.path}".')

        self.mode = "L"
        self.n_frames = _get_n_frames(n_frames)

        self.slices, max_size = [], 0
        for image_idx in range(len(self.hr_files)):
            image = Image.open(Path(self.path, self.hr_files[image_idx]))
            self.slices.append(1 if self.n_frames is None else image.n_frames // self.n_frames[0])
            max_size = max(max(image.size), max_size)

        self.val_idx = _get_val_idx(self.slices, val_split, split_seed)
        self.is_lr = max_size <= hr_res//lr_scale or (lr_scale == None)
        if self.is_lr: print("LR mode is enabled, dataset will load only unmodified low-resolution images.")
        self.crop_res = min(hr_res, max_size)

        self.hr_res = hr_res
        self.lr_scale = lr_scale
        self.crappifier = crappifier
        self.rotation = rotation
        self.transforms = transforms

    def __getitem__(self, idx, pp=False):
        if idx >= len(self): raise IndexError(f"Tried to retrieve invalid image. Index {idx} is not less than {len(self)} total image frame slices.")

        is_val = idx in self.val_idx or pp
        image_idx, idx = _get_image_idx(idx, self.slices)

        hr = _load_image(self.path, self.hr_files[image_idx], self.mode, self.n_frames[0] if self.n_frames is not None else None, self.slices[image_idx], idx)
        
        if self.is_lr:
            # Rotation and crappifier is disabled in lr mode
            return _ready_lr(hr, self.hr_res//self.lr_scale, self.transforms)

        return _gen_pair(hr, self.hr_res, self.lr_scale, False if is_val else self.rotation, self.crappifier, self.transforms, self.n_frames)
    
    def __len__(self):
        return sum(self.slices)
    
    def __repr__(self):
        return f'ImageDataset from path "{self.path}"\n{len(self.hr_files)} files with {len(self)} total frame slices\nLR mode {"enabled" if self.is_lr else "disabled"}'

    def _get_name(self, idx):
        image_idx, idx = _get_image_idx(idx, self.slices)
        return self.hr_files[image_idx].split('.')[0] + (f"_{idx}" if self.n_frames is not None else "")

class SlidingDataset(Dataset):
    def __init__(self, path : Path, hr_res : int = 512, lr_scale : int = 4, crappifier : Crappifier = Poisson(), overlap : int = 128, n_frames : list[int] = -1, stack : str = "TZ", extension : str = "czi", preload : bool = True, val_split : float = 0.1, rotation : bool = True, split_seed : int = 0, transforms : list[torch.nn.Module] = None):
        r"""Training dataset for loading high-resolution image tiles from image sheets and returning high-low-resolution pairs, the latter receiving crappification.

        Dataset used for image sheets (e.g. .czi files). For pre-tiled image files, use :class:`ImageDataset`.

        LR mode (dataset loads only unmodified low-resolution images for prediction) can be enabled by setting ``lr_scale`` = None and ``hr_res`` = LR resolution.

        Args:
            path (Path) : Path to folder containing high-resolution images. Can also be a str.

            hr_res (int) : Resolution of high-resolution images. Images larger than this will be downscaled to this resolution. Images smaller will be padded. Default is 512.

            lr_scale (int) : Downscaling factor for low-resolution images to simulate undersampling. Choose a power of 2 for best results. Default is 4.

            crappifier (Crappifier) : Crappifier for degrading low-resolution images to simulate undersampling. Not used in LR mode. Default is :class:`Poisson`.

            overlap (int) : Overlapping pixels between neighboring tiles to increase effective dataset size. Default is 128.

            n_frames (list[int]) : Amount of stacked frames per image, disregarding color channels. Can also be list of low-resolution and high-resolution stack amounts respectively. A value of -1 uses all stacked image frames. Default is -1.

            stack (str) : Multiframe stack handling mode, e.g "T" for time stack, "Z" for z dimension stack, "TZ" or "ZT" for both, determining flattenting order. Only applicable if loading from czi. Default is "TZ".

            extension (str) : File extension of images. Default is "czi".

            preload (bool) : Whether to preload images in memory (not VRAM) for faster dataloading. Default is True.

            val_split (float) : Proportion of images to be held out for evaluation/prediction. Default is 0.1.

            rotation (bool) : Whether to randomly rotate and/or flip images when loading data. Only used during training if applicable. Default is True.

            split_seed (int) : Seed for random train/evaluation data splitting. A value of None splits the last images as evaluation. Default is 0.

            transforms (list[nn.Module]) : Additional final data transforms to apply. Default is None.
        """
        super().__init__()
        self.path = Path(path) if type(path) is str else path
        if not path or not self.path.exists(): raise FileNotFoundError(f'Path "{self.path}" does not exist.')
        
        self.hr_files = _root_glob(f"*.{extension}", root_dir=self.path)
        if not len(self.hr_files) > 0: raise FileNotFoundError(f'No .{extension} files exist in path "{self.path}".')

        overlap = 0 if overlap is None else overlap
        if not hr_res > overlap: raise ValueError(f"hr_res must be greater than overlap. Given values are {hr_res} and {overlap} respectively.")
        self.stride = hr_res - overlap
        self.stack = stack.upper()
        self.mode = "L"
        self.n_frames = _get_n_frames(n_frames)
        
        self.preload = _preload(preload, [self.path], [self.hr_files], self.mode, self.stack)

        self.tiles, self.slices = [], []
        for image_idx in range(len(self.hr_files)):
            image = self.preload[image_idx] if self.preload else _load_sheet(self.path, self.hr_files[image_idx], self.stack, self.mode)
            tiles_x, tiles_y = _n_tiles(image, hr_res, self.stride)
            self.tiles.append(tiles_x * tiles_y)
            self.slices.append(1 if self.n_frames is None else image.shape[0] // self.n_frames[0])

        self.val_idx = _get_val_idx(self.slices, val_split, split_seed, self.tiles)
        self.is_lr = (lr_scale == None)
        if self.is_lr: print("LR mode is enabled, dataset will load only unmodified low-resolution images.")
        self.crop_res = hr_res

        self.hr_res = hr_res
        self.lr_scale = lr_scale
        self.crappifier = crappifier
        self.rotation = rotation
        self.transforms = transforms
    
    def __getitem__(self, idx, pp=False):
        if idx >= len(self): raise IndexError(f"Tried to retrieve invalid image. Index {idx} is not less than {len(self)} total image frame slices.")

        is_val = idx in self.val_idx or pp
        image_idx, idx = _get_image_idx(idx, self.slices, self.tiles)

        hr = _sliding_window(self.preload[image_idx] if self.preload else _load_sheet(self.path, self.hr_files[image_idx], self.stack, self.mode), self.hr_res, self.stride, self.n_frames[0] if self.n_frames is not None else None, self.slices[image_idx], idx)

        if self.is_lr:
            return _ready_lr(hr, self.hr_res, self.transforms)

        return _gen_pair(hr, self.hr_res, self.lr_scale, False if is_val else self.rotation, self.crappifier, self.transforms, self.n_frames)
    
    def __len__(self):
        return sum([self.tiles[idx] * self.slices[idx] for idx in range(len(self.hr_files))])
    
    def __repr__(self):
        return f'SlidingDataset from path "{self.path}"\n{len(self.hr_files)} files with {len(self)} total frame slices\nLR mode {"enabled" if self.is_lr else "disabled"}'
    
    def _get_name(self, idx):
        image_idx, idx = _get_image_idx(idx, self.slices, self.tiles)
        return f"{self.hr_files[image_idx].split('.')[0]}_{idx}"

class PairedImageDataset(Dataset):
    def __init__(self, hr_path : Path, lr_path : Path, hr_res : int = 512, lr_scale : int = 4, n_frames : list[int] = -1, extension : str = "tif", val_split : float = 1, rotation : bool = True, split_seed : int = None, transforms : list[torch.nn.Module] = None):
        r"""Testing dataset for loading paired high-low-resolution images without using crappification. Can also be used for approximating :class:`Crappifier` parameters.

        Args:
            hr_path (Path) : Path to folder containing high-resolution images. Can also be a str.

            lr_path (Path) : Path to folder containing low-resolution images. Can also be a str.

            hr_res (int) : Resolution of high-resolution images. Images larger than this will be downscaled to this resolution. Images smaller will be padded. Default is 512.

            lr_scale (int) : Downscaling factor for low-resolution images to simulate undersampling. Choose a power of 2 for best results. Default is 4.

            n_frames (list[int]) : Amount of stacked frames per image, disregarding color channels. Can also be list of low-resolution and high-resolution stack amounts respectively. A value of -1 uses all stacked image frames. Default is -1.

            extension (str) : File extension of images. Default is "tif".

            val_split (float) : Proportion of images to be held out for evaluation/prediction. Default is 1.

            rotation (bool) : Whether to randomly rotate and/or flip images when loading data. Only used during training if applicable. Default is True.

            split_seed (int) : Seed for random train/evaluation data splitting. A value of None splits the last images as evaluation. Default is None.

            transforms (list[nn.Module]) : Additional final data transforms to apply. Default is None.
        """
        super().__init__()
        self.hr_path = Path(hr_path) if type(hr_path) is str else hr_path
        self.lr_path = Path(lr_path) if type(lr_path) is str else lr_path
        for path in [self.hr_path, self.lr_path]:
            if not path or not path.exists(): raise FileNotFoundError(f'Path "{path}" does not exist.')
        if self.hr_path == self.lr_path: warnings.warn("hr_path is equal to lr_path! Consider using ImageDataset instead.", stacklevel=2)

        self.hr_files = _root_glob(f"*.{extension}", root_dir=self.hr_path)
        self.lr_files = _root_glob(f"*.{extension}", root_dir=self.lr_path)
        for files, path in zip([self.hr_files, self.lr_files], [self.hr_path, self.lr_path]):
            if not len(files) > 0: raise FileNotFoundError(f'No .{extension} files exist in path "{path}".')
        if len(self.hr_files) != len(self.lr_files): raise FileNotFoundError(f"Mismatch between amounts of high-low-resolution images. Found {len(self.hr_files)} high-resolution and {len(self.lr_files)} low-resolution images.")

        self.mode = "L"
        self.n_frames = _get_n_frames(n_frames)

        self.slices, max_size = [], 0
        for image_idx in range(len(self.hr_files)):
            image = Image.open(Path(self.hr_path, self.hr_files[image_idx]))
            self.slices.append(1 if self.n_frames is None else image.n_frames // self.n_frames[0])
            max_size = max(max(image.size), max_size)

        self.val_idx = _get_val_idx(self.slices, val_split, split_seed)
        self.is_lr = False
        self.crop_res = min(hr_res, max_size)

        self.hr_res = hr_res
        self.lr_scale = lr_scale
        self.rotation = rotation
        self.transforms = transforms
    
    def __getitem__(self, idx, pp=False):
        if idx >= len(self): raise IndexError(f"Tried to retrieve invalid image. Index {idx} is not less than {len(self)} total image frame slices.")

        is_val = idx in self.val_idx or pp
        image_idx, idx = _get_image_idx(idx, self.slices)

        hr = _load_image(self.hr_path, self.hr_files[image_idx], self.mode, self.n_frames[0] if self.n_frames is not None else None, self.slices[image_idx], idx)
        lr = _load_image(self.lr_path, self.lr_files[image_idx], self.mode, self.n_frames[0] if self.n_frames is not None else None, self.slices[image_idx], idx)

        return _transform_pair(hr, lr, self.hr_res, self.hr_res//self.lr_scale, False if is_val else self.rotation, self.transforms, self.n_frames)
    
    def __len__(self):
        return sum(self.slices)
    
    def __repr__(self):
        return f'PairedImageDataset from paths "{self.hr_path}" and "{self.lr_path}"\n{len(self.hr_files)} paired files with {len(self)} total frame slices'

    def _get_name(self, idx):
        image_idx, idx = _get_image_idx(idx, self.slices)
        return self.lr_files[image_idx].split('.')[0] + (f"_{idx}" if self.n_frames is not None else "")

class PairedSlidingDataset(Dataset):
    def __init__(self, hr_path : Path, lr_path : Path, hr_res : int = 512, lr_scale : int = 4, overlap : int = 128, n_frames : list[int] = -1, stack : str = "TZ", extension : str = "czi", preload : bool = True, val_split : float = 1, rotation : bool = True, split_seed : int = None, transforms : list[torch.nn.Module] = None):
        r"""Testing dataset for loading high-low-resolution image tiles from image sheets without crappification. Can also be used for approximating :class:`Crappifier` parameters.

        Dataset used for image sheets (e.g. .czi files). For pre-tiled image files, use :class:`ImageDataset`.

        Args:
            hr_path (Path) : Path to folder containing high-resolution images. Can also be a str.

            lr_path (Path) : Path to folder containing low-resolution images. Can also be a str.

            hr_res (int) : Resolution of high-resolution images. Images larger than this will be downscaled to this resolution. Images smaller will be padded. Default is 512.

            lr_scale (int) : Downscaling factor for low-resolution images to simulate undersampling. Choose a power of 2 for best results. Default is 4.

            overlap (int) : Overlapping pixels between neighboring tiles to increase effective dataset size. Default is 128.

            n_frames (list[int]) : Amount of stacked frames per image, disregarding color channels. Can also be list of low-resolution and high-resolution stack amounts respectively. A value of -1 uses all stacked image frames. Default is -1.

            stack (str) : Multiframe stack handling mode, e.g "T" for time stack, "Z" for z dimension stack, "TZ" or "ZT" for both, determining flattenting order. Only applicable if loading from czi. Default is "TZ".

            extension (str) : File extension of images. Default is "czi".

            preload (bool) : Whether to preload images in memory (not VRAM) for faster dataloading. Default is True.

            val_split (float) : Proportion of images to be held out for evaluation/prediction. Default is 1.

            rotation (bool) : Whether to randomly rotate and/or flip images when loading data. Only used during training if applicable. Default is True.

            split_seed (int) : Seed for random train/evaluation data splitting. A value of None splits the last images as evaluation. Default is None.

            transforms (list[nn.Module]) : Additional final data transforms to apply. Default is None.
        """
        super().__init__()
        self.hr_path = Path(hr_path) if type(hr_path) is str else hr_path
        self.lr_path = Path(lr_path) if type(lr_path) is str else lr_path
        for path in [self.hr_path, self.lr_path]:
            if not path or not path.exists(): raise FileNotFoundError(f'Path "{path}" does not exist.')
        if self.hr_path == self.lr_path: warnings.warn("hr_path is equal to lr_path! Consider using SlidingDataset instead.", stacklevel=2)
        
        self.hr_files = _root_glob(f"*.{extension}", root_dir=self.hr_path)
        self.lr_files = _root_glob(f"*.{extension}", root_dir=self.lr_path)
        for files, path in zip([self.hr_files, self.lr_files], [self.hr_path, self.lr_path]):
            if not len(files) > 0: raise FileNotFoundError(f'No .{extension} files exist in path "{path}".')
        if len(self.hr_files) != len(self.lr_files): raise FileNotFoundError(f"Mismatch between amounts of high-low-resolution images. Found {len(self.hr_files)} high-resolution and {len(self.lr_files)} low-resolution images.")

        overlap = 0 if overlap is None else overlap
        if not hr_res > overlap: raise ValueError(f"hr_res must be greater than overlap. Given values are {hr_res} and {overlap} respectively.")
        self.stride = hr_res - overlap
        self.stack = stack.upper()
        self.mode = "L"
        self.n_frames = _get_n_frames(n_frames)

        self.preload = _preload(preload, [self.hr_path, self.lr_path], [self.hr_files, self.lr_files], self.mode, self.stack)

        self.tiles, self.slices = [], []
        for image_idx in range(len(self.hr_files)):
            image = self.preload[0][image_idx] if self.preload else _load_sheet(self.hr_path, self.hr_files[image_idx], self.stack, self.mode)
            tiles_x, tiles_y = _n_tiles(image, hr_res, self.stride)
            self.tiles.append(tiles_x * tiles_y)
            self.slices.append(1 if self.n_frames is None else image.shape[0] // self.n_frames[0])

        self.val_idx = _get_val_idx(self.slices, val_split, split_seed, self.tiles)
        self.is_lr = False
        self.crop_res = hr_res

        self.hr_res = hr_res
        self.lr_scale = lr_scale
        self.rotation = rotation
        self.transforms = transforms
    
    def __getitem__(self, idx, pp=False):
        if idx >= len(self): raise IndexError(f"Tried to retrieve invalid image. Index {idx} is not less than {len(self)} total image frame slices.")

        is_val = idx in self.val_idx or pp
        image_idx, idx = _get_image_idx(idx, self.slices, self.tiles)

        hr = _sliding_window(self.preload[0][image_idx] if self.preload else _load_sheet(self.hr_path, self.hr_files[image_idx], self.stack, self.mode), self.hr_res, self.stride, self.n_frames[0] if self.n_frames is not None else None, self.slices[image_idx], idx)
        lr = _sliding_window(self.preload[1][image_idx] if self.preload else _load_sheet(self.lr_path, self.lr_files[image_idx], self.stack, self.mode), self.hr_res//self.lr_scale, self.stride//self.lr_scale, self.n_frames[0] if self.n_frames is not None else None, self.slices[image_idx], idx)

        return _transform_pair(hr, lr, self.hr_res, self.hr_res//self.lr_scale, False if is_val else self.rotation, self.transforms, self.n_frames)
    
    def __len__(self):
        return sum([self.tiles[idx] * self.slices[idx] for idx in range(len(self.hr_files))])
    
    def __repr__(self):
        return f'PairedSlidingDataset from paths "{self.hr_path}" and "{self.lr_path}"\n{len(self.hr_files)} paired files with {len(self)} total frame slices'
    
    def _get_name(self, idx):
        image_idx, idx = _get_image_idx(idx, self.slices, self.tiles)
        return f"{self.lr_files[image_idx].split('.')[0]}_{idx}"

def preprocess_dataset(dataset : Dataset, preprocess_hr : bool = False, out_dir : str = "preprocess"):
    r"""Saves processed frame slices from a given dataset, including processes such as crappification or cropping/padding.
    
    Args:
        dataset (Dataset) : Dataset to load images from. Preprocessing is determined by the dataloading procedure as specified in dataset arguments. Rotation is disabled.

        preprocess_hr (bool) : Whether to save preprocessed high-resolution images in addition to preprocessed low-resolution images. Default is False.

        out_dir (str) : Directory to save preprocessed images. Default is "preprocess".
    """
    os.makedirs(f"{out_dir}/lr", exist_ok=True)
    if preprocess_hr:
        os.makedirs(f"{out_dir}/hr", exist_ok=True)

    progress = tqdm(range(len(dataset)))
    for idx in progress:
        hr, lr = dataset.__getitem__(idx, pp=True)
        hr, lr = np.asarray(hr, dtype=np.uint8), np.asarray(lr, dtype=np.uint8)
        
        tifffile.imwrite(f"{out_dir}/lr/{dataset._get_name(idx)}.tif", lr)
        if preprocess_hr:
            tifffile.imwrite(f"{out_dir}/hr/{dataset._get_name(idx)}.tif", hr)

# TODO: crappify_images

def _gen_pair(hr, hr_res, lr_scale, rotation, crappifier, transforms, n_frames):
    r"""Creates training ready pair of images from a single high-resolution image.
    """
    hr = _square_crop(hr, hr_res)
    hr = _pad_image(hr, hr_res)
    
    # Set random rotation and flip in xy axis
    if rotation:
        hr = np.rot90(hr, axes=(1,2)) if bool(random.getrandbits(1)) else hr
        hr = np.flip(hr, axis=random.choice((1,2,(1,2))))

    # Crappification
    lr = np.stack([Image.fromarray(channel).resize(([hr_res//lr_scale]*2), Image.Resampling.BILINEAR) for channel in hr])
    if crappifier is not None:
        # Allows either Crappifier or nn.Module (or any callable function) to be used as a crappifier
        lr = crappifier.crappify(lr) if issubclass(type(crappifier), Crappifier) else crappifier(lr)
        lr = np.clip(lr.round(), 0, 255)

    hr = _slice_center(hr, n_frames)

    return _tensor_ready(hr, transforms), _tensor_ready(lr, transforms)

def _transform_pair(hr, lr, hr_res, lr_res, rotation, transforms, n_frames):
    r"""Same as _gen_pair, but uses paired high-low-resolution images with no crappifier.
    """
    hr = _square_crop(hr, hr_res)
    lr = _square_crop(lr, lr_res)

    hr = _pad_image(hr, hr_res)
    lr = _pad_image(lr, lr_res)

    if rotation:
        choice = bool(random.getrandbits(1)), random.choice((0,1,(0,1)))

        hr, lr = np.rot90(hr, axes=(1,2)) if choice[0] else hr, np.rot90(lr, axes=(1,2)) if choice[0] else lr
        hr, lr = np.flip(hr, axis=choice[1]), np.flip(lr, axis=choice[1])
    
    hr = _slice_center(hr, n_frames)

    return _tensor_ready(hr, transforms), _tensor_ready(lr, transforms)

def _ready_lr(lr, lr_res, transforms):
    r"""Processes lr images for prediction. Same as transform pair but without rotation.
    """
    lr = _square_crop(lr, lr_res)
    lr = _pad_image(lr, lr_res)

    return _tensor_ready(lr, transforms)

def _tensor_ready(image, transforms):
    image = torch.tensor(image.copy(), dtype=torch.float)

    # Additional nn.Module user transforms
    if transforms is not None:
        for transform in transforms:
            image = transform(image)
    
    return image

def _square_crop(image, max_res):
    height, width = image.shape[-2:]

    if [height, width] == [max_res]*2:
        return image
    
    size = min(height, width, max_res)
    start_x = (height - size) // 2
    start_y = (width - size) // 2
    
    return image[:, start_x:start_x + size, start_y:start_y + size]

def _pad_image(image, res):
    if image.shape[-1] < res:
        return np.stack([np.pad(channel, pad_width=[[0,res-image.shape[-1]]]*2, mode="reflect") for channel in image])
    return image

def _preload(preload, path, files, mode, stack):
    if not preload:
        return False
    
    size = sum([sum([os.stat(Path(idx_path, file)).st_size for file in idx_files]) for idx_path, idx_files in zip(path, files)]) / 10**9
    memory = psutil.virtual_memory().available / 10**9
    if size > memory:
        warnings.warn(f"Total dataset size {size:.2f}GB is greater than available memory of {memory:.2f}GB. Consider disabling preloading to avoid potential slowdowns.", stacklevel=2)

    print(f"Preloading {sum([len(item) for item in files])} images into memory...")
    preload = [[_load_sheet(idx_path, file, stack, mode) for file in idx_files] for idx_path, idx_files in zip(path, files)]
    return preload[0] if len(preload) == 1 else preload

def _load_image(path, file, mode, n_frames, slices, idx):
    hr = Image.open(Path(path, file))

    hr = _frame_channel(hr, mode)
    return _slice_image(hr, n_frames, slices, idx)

def _load_sheet(path, file, stack, mode):
    extension = file.split(".")[-1]
    if extension == "czi":
        # Retrieve same channel information from czi regardless of channel order
        image = czifile.CziFile(Path(path, file))
        out_axes = "TZCXY"

        # Get info only from out_axes axes
        slice_idx, slice_axes = [], []
        for axis in image.axes:
            if axis not in out_axes:
                slice_idx.append(0)
            else:
                slice_idx.append(slice(None))
                slice_axes.append(axis)
        image = image.asarray()[tuple(slice_idx)]

        # Rearrange axes to out_axes order
        axes_idx = [out_axes.rfind(axis) for axis in slice_axes]
        image = np.moveaxis(image, range(len(image.shape)), axes_idx)

        # TODO: Assumes images must have all out_axes?
        # Disregard additional channels and/or reorder
        if mode == "L":
            image = np.mean(image, axis=2)
        if stack == "T":
            image = image[:,0]
        elif stack == "Z":
            image = image[0]
        elif stack == "ZT":
            image = np.moveaxis(image, 0, 1)
        elif stack == "TZ":
            pass
        else:
            raise ValueError(f"Stack type {stack} is not valid.")
        
        # Flatten channel dimensions
        image = np.reshape(image, [-1, image.shape[-2], image.shape[-1]])
        if image.max() != 0:
            image = image / (image.max() / 255)
        return image.astype(np.uint8)
    else:
        image = Image.open(Path(path, file))
        return _frame_channel(image, mode)

def _sliding_window(image, size, stride, n_frames, n_slices, idx):
    tiles_x, tiles_y = _n_tiles(image, size, stride)
    tile_idx = idx // n_slices

    start_x = tile_idx // tiles_y * stride
    start_y = tile_idx % tiles_y * stride

    image = image[..., start_x:start_x + size, start_y:start_y + size]

    return _slice_image(image, n_frames, n_slices, idx)

def _frame_channel(image, mode):
    # Create frame dimension
    if image.n_frames > 1:
        image = np.stack([np.asarray(_seek_channel(image, frame).convert(mode), dtype=np.uint8) for frame in range(image.n_frames)])
    else:
        image = np.asarray(image.convert(mode), dtype=np.uint8)[np.newaxis, :, :]
    
    return image

def _slice_image(image, n_frames, n_slices, idx):
    if n_frames == None:
        return image
    
    # Only need to calculate residual idx of specific tile
    idx = idx % n_slices
    idx *= n_frames
    return image[idx:idx+n_frames]

def _slice_center(image, n_frames):
    if n_frames is None or n_frames[0] == n_frames[1] or n_frames[1] > image.shape[-3]:
        return image

    center = image.shape[-3] // 2
    half = n_frames[1] // 2
    if n_frames[1] % 2 == 0:
        return image[...,center - half:center + half,:,:]
    else:
        return image[...,center - half:center + half + 1,:,:]

def _seek_channel(image, idx):
    # Because we can't have nice things...
    image.seek(idx)
    return image

def _root_glob(search, root_dir):
    files = glob.glob(f"{root_dir}/{search}")
    return sorted([item.split("/")[-1] for item in files])

def _n_tiles(image, size, stride):
    x, y = image.shape[-2:]

    tiles_x = max(0, (x - size) // stride + 1)
    tiles_y = max(0, (y - size) // stride + 1)
    return tiles_x, tiles_y

def _get_n_frames(n_frames):
    if n_frames in [None, -1, [-1]]:
        return None
    
    n_frames = _force_list(n_frames)
    n_frames = n_frames*2 if len(n_frames) == 1 else n_frames
    return n_frames # [in (low-resolution), out (high-resolution)]

def _get_image_idx(idx, slices, tiles=None):
    tiles = [1]*len(slices) if tiles is None else tiles

    image_idx = 0
    for slice, tile in zip(slices, tiles):
        if idx < slice * tile:
            return image_idx, idx
        else:
            idx -= slice * tile
            image_idx += 1

def _get_val_idx(slices, split, seed, tiles=None):
    # Tiles are individual images
    if tiles is not None:
        tile_slices = []
        for slice, tile in zip(slices, tiles):
            tile_slices.extend([slice]*tile)
        slices = tile_slices

    # Select random image indexes as validation
    val_slices = list(range(len(slices)))
    if seed is not None and split < 1:
        np.random.seed(seed)
        np.random.shuffle(val_slices)
    val_slices = val_slices[-max(1,int(split*len(slices))):]
    
    # Convert image indexes to frame indexes for dataloading
    val_idx, idx = [], 0
    for slice_idx, slice in enumerate(slices):
        if slice_idx in val_slices:
            val_idx.extend(range(idx,idx+slice))
        idx += slice

    return val_idx

def _invert_idx(idx, idx_len):
    idx_range = np.arange(idx_len)
    inverse = np.logical_not(np.isin(idx_range, idx))
    return idx_range[inverse]

class _RandomIterIdx:
    def __init__(self, idx, seed=False):
        self.idx = idx
        self.seed = seed
        
    def __iter__(self):
        random_idx = self.idx.copy()
        if self.seed:
            np.random.seed(0)
            np.random.shuffle(random_idx)
        else:
            random.shuffle(random_idx)
        yield from random_idx
    
    def __len__(self):
        return len(self.idx)



import numpy as np
from abc import ABC, abstractmethod
from skimage.util import random_noise

class Crappifier(ABC):
    r"""Crappifier base class for custom crappifiers. Override the :meth:`crappify` method for logic.
    """
    @abstractmethod
    def crappify(self, image : np.ndarray):
        r"""An abstract function for degrading a low resolution image to simulate undersampling.

        This method is not responsible for downscaling the image, only for injecting noise.

        Args:
            image (np.ndarray) : Low resolution image to crappify.

        Returns:
            crap (np.ndarray) : The low resolution image, only now has it been crappified.
        """
        raise NotImplementedError('"crappify" method not implemented.')
    
    def __call__(self, image : np.ndarray):
        return self.crappify(image)

class MultiCrappifier(Crappifier):
    def __init__(self, *args : list[Crappifier], clip : bool = True):
        r"""Chains multiple crappifiers sequentially for degrading low resolution images.

        Args:
            args (Crappifier) : Crappifiers to be applied in order from first to last.

            clip (bool) : Clip values to image range between each crappifier step. Default is True.
        """
        self.crappifiers = args
        self.clip = clip
    
    def crappify(self, image : np.ndarray):
        for crappifier in self.crappifiers:
            image = crappifier.crappify(image)
            if self.clip:
                image = np.clip(image, 0, 255)
        return image

class AdditiveGaussian(Crappifier):
    def __init__(self, intensity : float = 13, gain : float = 0, spread : float = 0):
        r"""Crappifier using additive Gaussian noise sampling (normally distributed noise). Adds additive Gaussian noise to a low resolution image.

        Approximates :class:`Poisson` noise at high samples.

        Args:
            intensity (float) : Standard deviation of Gaussian distribution. Higher values will introduce more noise to the image. Default is 13.

            gain (float) : Mean of Gaussian distribution. Higher or lower values will raise the mean image value higher or lower respectively. Default is 0.

            spread (float) : Standard deviation of crappifier intensity for training on a range of crappifications. Default is 0.
        """
        self.intensity = intensity
        self.gain = gain
        self.spread = spread

    def crappify(self, image : np.ndarray):
        intensity = max(np.random.normal(self.intensity, self.spread), 0) if self.spread > 0 else self.intensity
        return image + np.random.normal(self.gain, intensity, image.shape)
    
class Poisson(Crappifier):
    def __init__(self, intensity : float = 1, gain : float = 0, spread : float = 0):
        r"""Crappifier using Poisson noise sampling (shot noise). Adds Poisson noise to a low resolution image.

        Args:
            intensity (float) : Interpolated mix of generated Poisson image. 1 is standard distribution, 0 is none, larger values amplify noise. Default is 1.

            gain (float) : Value gain added to the output. Default is 0.

            spread (float) : Standard deviation of crappifier intensity for training on a range of crappifications. Default is 0.
        """
        self.intensity = intensity
        self.gain = gain
        self.spread = spread
        
    def crappify(self, image : np.ndarray):
        return self._interpolate(image, np.random.poisson(np.clip(image, 0, np.inf))) + self.gain
    
    def _interpolate(self, x, y):
        intensity = max(np.random.normal(self.intensity, self.spread), 0) if self.spread > 0 else self.intensity
        return x * (1 - intensity) + y * intensity
    
class SaltPepper(Crappifier):
    def __init__(self, intensity : float = 0.5, gain : float = 0, spread : float = 0):
        r"""Crappifier using salt and pepper noise (full value addition/deletion). Adds salt and pepper noise to a low resolution image.

        Args:
            intensity (float) : Percent of values to replace with salt and pepper noise. Default is 0.5.

            gain (float) : Value gain added to the output. Default is 0.

            spread (float) : Standard deviation of crappifier intensity for training on a range of crappifications. Default is 0.
        """
        self.intensity = intensity / 100
        self.gain = gain
        self.spread = spread
        
    def crappify(self, image : np.ndarray):
        intensity = max(np.random.normal(self.intensity, self.spread), 0) if self.spread > 0 else self.intensity
        return random_noise(np.clip(image, 0, 255)/255, mode="s&p", amount=intensity) * 255 + self.gain



import napari, warnings
import numpy as np
from qtpy.QtWidgets import QWidget, QVBoxLayout
from qtpy.QtCore import QObject, QThread, Signal, Qt
from magicgui.widgets import Container, ComboBox, PushButton, FileEdit, TextEdit, create_widget
from torch.nn import MSELoss
from enum import Enum
from PIL import Image
from ..__main__ import pssr_head
from ._util import ObjectEdit, SignalWrapper
from ..crappifiers import AdditiveGaussian, Poisson, SaltPepper
from ..models import ResUNet, RDResUNet, SwinIR
from ..data import ImageDataset, SlidingDataset, PairedImageDataset, PairedSlidingDataset
from ..util import SSIMLoss

try:
    from matplotlib.backends.backend_qtagg import FigureCanvasQTAgg
    from matplotlib.figure import Figure
    USE_PLOT = True
except:
    warnings.warn("matplotlib could not be imported, plotting features will be disabled.", stacklevel=2)
    USE_PLOT = False

class Status(Enum):
    IDLE_TRAIN = "Train Model"
    IDLE_PREDICT = "Predict Images"
    LOADING = "Loading..."
    PROGRESS_TRAIN = "Cancel Training"
    PROGRESS_PREDICT = "Cancel Predicting"

class PSSRWidget(QWidget):
    def __init__(self, is_train : bool):
        super().__init__()
        self.viewer = napari.current_viewer()

        self.model = ObjectEdit("Model", [ResUNet, RDResUNet, SwinIR])
        self.dataset = ObjectEdit("Dataset", [ImageDataset, SlidingDataset, PairedImageDataset, PairedSlidingDataset], hide_crappifier=not is_train)

        self.device = ComboBox(name="Device", choices=["cuda", "cpu"])
        self.model_path = FileEdit(name="Model Path")
        self.batch_size = create_widget(value=16, name="Batch Size")

        if is_train:
            self.epochs = create_widget(value=10, name="Epochs")
            self.lr = create_widget(value=0.001, name="Learning Rate", options=dict(step=1e-5))
            self.patience = create_widget(value=3, name="Patience")
            self.loss_fn = ComboBox(name="Loss Function", choices=["MS-SSIM", "SSIM", "MSE"])
            self.checkpoint = create_widget(value=False, name="Save Checkpoints")
            self.losses = create_widget(value=False, name="Save Losses")
            self.resume = create_widget(value=False, name="Load Checkpoint")

            self.resume.changed.connect(lambda : setattr(self.model_path, "visible", self.resume.value))
            self.model_path.visible = self.resume.value

        self.params = Container()
        self.params.append(self.device)
        if is_train:
            self.params.append(self.epochs)
            self.params.append(self.batch_size)
            self.params.append(self.lr)
            self.params.append(self.patience)
            self.params.append(self.loss_fn)
            self.params.append(self.checkpoint)
            self.params.append(self.losses)
            self.params.append(self.resume)
        else:
            pass
        self.params.append(self.model_path)

        self.trigger = PushButton(text=Status.IDLE_TRAIN.value if is_train else Status.IDLE_PREDICT.value)
        self.trigger.changed.connect(lambda : self.process_wrapper(is_train))

        self.console = TextEdit(value="")
        self.console.read_only = True
        self.console.hide()
        self.err_len = 0

        layout = QVBoxLayout()
        layout.addWidget(self.model)
        layout.addWidget(self.dataset)
        layout.addWidget(self.params.native)
        layout.addWidget(self.trigger.native)
        layout.addWidget(self.console.native)
        layout.setAlignment(Qt.AlignTop)
        self.setLayout(layout)

        self.hide = [self.model, self.dataset, self.params]
        self.plot = None
    
    def process_wrapper(self, train : bool):
        # Functions as cancel button when running
        if self.trigger.text in [Status.LOADING.value, Status.PROGRESS_TRAIN.value, Status.PROGRESS_PREDICT.value]:
            self.worker.abort = True
            return

        if train:
            self.worker = TrainProcess(
                model=self.model.object,
                dataset=self.dataset.object,
                device=self.device.value,
                epochs=self.epochs.value,
                batch_size=self.batch_size.value,
                lr=self.lr.value,
                patience=self.patience.value,
                loss_fn=self.loss_fn.value,
                checkpoint=self.checkpoint.value,
                losses=self.losses.value,
                model_path=self.model_path.value if self.resume.value else None,
            )
        else:
            self.worker = PredictProcess(
                model=self.model.object,
                dataset=self.dataset.object,
                device=self.device.value,
                model_path=self.model_path.value,
            )
        self.worker.stage.connect(lambda x : setattr(self.trigger, "text", x))
        self.worker.monitor.connect(lambda x : setattr(eval(f"self.viewer.layers['{x[0]}']", {"self":self}), "data", x[1]))
        self.worker.finished.connect(self._close_thread)
        self.worker.error.connect(lambda x : self._catch_error(x))

        # Capture sys.stdout to "console"
        self.capture = SignalWrapper(self.worker.run)
        self.capture.out.connect(lambda x : self._write_console(x))
        self.capture.err.connect(lambda x : self._write_console(x, err=True))
        self.console.value = ""

        self.thread = QThread()
        self.capture.moveToThread(self.thread)
        self.thread.started.connect(self.capture.capture)

        # Adjust UI
        for widget in self.hide:
            widget.hide()
        self.model.collapse.collapse()
        self.dataset.collapse.collapse()
        self.console.show()

        if USE_PLOT:
            if self.plot is not None:
                self.layout().removeWidget(self.plot)
                self.plot.hide()
                self.plot = None
            
            if train:
                self.plot = LossPlot(size=(6,4))
                self.worker.loss.connect(lambda x : self.plot.add_point(x))
                self.layout().addWidget(self.plot)
            else:
                self.plot = MetricsPlot(size=(6,4))
                self.worker.metrics.connect(lambda x : self.plot.show_metrics(x))
                self.layout().addWidget(self.plot)
                self.plot.hide()
        
        for name in ["LR", "PSSR", "HR"] if train else []:
            try:
                self.viewer.layers.remove(name)
            except:
                pass
            
            self.viewer.add_image(np.zeros(shape=[1]+[self.dataset.arguments["hr_res"]]*2, dtype=np.uint8), name=name)

        self.thread.start()
    
    def _write_console(self, line : str, err : bool = False):
        if len(line.strip()) > 0:
            if err:
                # Remove previous stderr lines
                self.console.value = "\n".join(self.console.value.strip().split("\n")[self.err_len:])
                self.err_len = len(line.split("\n"))
                line = line.strip()
            elif self.err_len > 0:
                self.err_len = 0

            if line[-1] == "\n":
                line = "\n"+line.strip()

            self.console.value = line+"\n"+self.console.value

    def _close_thread(self):
        self.thread.quit()
        self.thread.wait()

        for widget in self.hide:
            widget.show()
    
    def _catch_error(self, error : Exception):
        self._close_thread()
        self._write_console(repr(error))
        raise error

class TrainProcess(QObject):

    stage = Signal(str)

    monitor = Signal(list)

    finished = Signal(bool)
    error = Signal(Exception)

    if USE_PLOT:
        loss = Signal(float)

    def __init__(self, model, dataset, device, epochs, batch_size, lr, patience, loss_fn, checkpoint, losses, model_path):
        super().__init__()
        self.model = model
        self.dataset = dataset
        self.device = device
        self.epochs = epochs
        self.batch_size = batch_size
        self.lr = lr
        self.patience = patience
        self.loss_fn = loss_fn
        self.checkpoint = checkpoint
        self.losses = losses
        self.model_path = model_path

        if loss_fn == "MS-SSIM":
            self.loss_fn = SSIMLoss()
        elif loss_fn == "SSIM":
            self.loss_fn = SSIMLoss(ms=False)
        else:
            self.loss_fn = MSELoss()

        self.abort = False

    def run(self):
        try:
            pssr_head(
                train=True,
                model=self.model,
                dataset=self.dataset,
                device=self.device,
                epochs=self.epochs,
                batch_size=self.batch_size,
                lr=self.lr,
                patience=self.patience,
                loss_fn=self.loss_fn,
                checkpoint=self.checkpoint,
                losses=self.losses,
                model_path=self.model_path,
                callbacks=[self._abort_callback, self._viewer_callback]+([self._plot_callback] if USE_PLOT else []),
                stage=self.stage,
            )

            self.finished.emit(True)

        except Exception as error:
            self.error.emit(error)

        # Triggers before exception
        finally:
            self.stage.emit(Status.IDLE_TRAIN.value)
    
    def _abort_callback(self):
        if self.abort:
            raise InterruptedError("Model training cancelled by user")

    def _viewer_callback(self, train_locals):
        batch_idx, log_frequency, progress = train_locals["batch_idx"], train_locals["log_frequency"], train_locals["progress"]
        if batch_idx % log_frequency == 0 or batch_idx == len(progress) - 1:
            data = train_locals["last_full"] if batch_idx == len(progress) - 1 else [train_locals["lr"].cpu(), train_locals["hr_hat"].cpu(), train_locals["hr"].cpu()]
            lr, hr_hat, hr = [image.detach().numpy().astype(np.uint8) for image in data]
            channels = max([lr.shape[1], hr_hat.shape[1], hr.shape[1]])

            for name, batched in zip(["LR", "PSSR", "HR"], [lr, hr_hat, hr]):
                if name == "LR":
                    batched = np.stack([[Image.fromarray(channel).resize(hr.shape[-2:], Image.Resampling.NEAREST) for channel in image] for image in lr])

                collage = self._collage_images(batched)
                if collage.shape[0] == 1 and collage.shape[0] < channels:
                    collage = np.repeat(collage, channels, axis=0)

                self.monitor.emit([name, collage])

    if USE_PLOT:
        def _plot_callback(self, train_locals):
            batch_idx, log_frequency, progress = train_locals["batch_idx"], train_locals["log_frequency"], train_locals["progress"]
            if batch_idx % log_frequency == 0 or batch_idx == len(progress) - 1:
                self.loss.emit(train_locals["loss"].item())
    
    def _collage_images(self, batched):
        n_rows = int(np.sqrt(batched.shape[0]))
        n_cols = batched.shape[0]//n_rows

        image_size = batched.shape[-1]
        collage = np.zeros([batched.shape[1], n_rows*image_size, n_cols*image_size])

        for idx in range(n_rows*n_cols):
            row = idx // n_cols
            col = idx % n_cols
            collage[:, row*image_size:(row+1)*image_size, col*image_size:(col+1)*image_size] = batched[idx]
        return collage

class PredictProcess(QObject):

    stage = Signal(str)

    monitor = Signal(list)

    finished = Signal(bool)
    error = Signal(Exception)

    if USE_PLOT:
        metrics = Signal(list)

    def __init__(self, model, dataset, device, model_path):
        super().__init__()
        self.model = model
        self.dataset = dataset
        self.device = device
        self.model_path = model_path

        self.abort = False

    def run(self):
        try:
            pssr_head(
                train=False,
                model=self.model,
                dataset=self.dataset,
                device=self.device,
                epochs=None,
                batch_size=None,
                lr=None,
                patience=None,
                loss_fn=None,
                checkpoint=None,
                losses=None,
                model_path=self.model_path,
                callbacks=[self._abort_callback],
                stage=self.stage,
                metrics=self.metrics,
            )

            self.finished.emit(True)

        except Exception as error:
            self.error.emit(error)

        # Triggers before exception
        finally:
            self.stage.emit(Status.IDLE_PREDICT.value)
    
    def _abort_callback(self):
        if self.abort:
            raise InterruptedError("Prediction cancelled by user")

if USE_PLOT:
    class LossPlot(FigureCanvasQTAgg):
        def __init__(self, size):
            fig = Figure(figsize=size)
            super().__init__(fig)

            self.ax = fig.add_subplot()
            self.ax.set_title("Training Loss")
            self.ax.tick_params(
                axis="x",
                which="both",
                bottom=False,
                top=False,
                labelbottom=False,
            )

            fig.set_tight_layout(True)

            self.pos = -1
            self.last = None
        
        def add_point(self, value):
            # Skip plotting ugly first batch
            if self.pos > 0:
                self.ax.plot([self.pos-1, self.pos], [self.last, value], c="blue")
                self.draw()

            self.pos += 1
            self.last = value
    
    class MetricsPlot(FigureCanvasQTAgg):
        def __init__(self, size):
            fig = Figure(figsize=size)
            super().__init__(fig)

            self.ax_psnr = fig.add_subplot(121)
            self.ax_psnr.set_title("PSNR")

            self.ax_ssim = fig.add_subplot(122)
            self.ax_ssim.set_title("SSIM")

            fig.set_tight_layout(True)
        
        def show_metrics(self, data):
            self._boxplot(self.ax_psnr, data[0])
            self._boxplot(self.ax_ssim, data[1])
            self.show()
        
        def _boxplot(self, ax, data):
            ax.boxplot(data, showfliers=False)
            x = np.random.normal(1, 0.02, size=len(data))
            ax.plot(x, data, ".", alpha=0.5)

class TrainWidget(PSSRWidget):
    def __init__(self):
        super().__init__(is_train=True)

class PredictWidget(PSSRWidget):
    def __init__(self):
        super().__init__(is_train=False)



import torch
import torch.nn as nn
import torch.nn.functional as F
from ._blocks import PSP_Pooling, Reconstruction, get_resblock
from ..util import _force_list

class ResUNet(nn.Module):
    def __init__(
            self,
            channels : list[int] = 1,
            hidden : list[int] = [64, 128, 256, 512, 1024],
            scale : int = 4,
            depth : int = 3,
            dilations : list[list[int]] = None,
            pool_sizes : list[int] = None,
            encoder_pool : bool = False,
        ):
        r"""A Residual UNet as detailed in Zhang et al., 2017 with an additional image upscaling block.
        If ``dilations`` is provided, instead use a Atrous Residual UNet as detailed in Diakogiannis et al., 2019.

        Channel sizes hidden[0] (and hidden[-1] if encoder_pool is True) must be divisible by pool_sizes if provided.

        Args:
            channels (list[int]) : Number of channels in image data. Can also be a list of in channels (low-resolution) and out channels (high-resolution) respectively.

            hidden (list[int]) : Elementwise list of channels per residual block controlling width and length of model.

            scale (int) : Upscaling factor for predictions. Choose a power of 2 for best results. Default is 4.

            depth (int) : Number of hidden layers per residual block. Default is 3.

            dilations (list[list[int]]) : List of dilation values per layer. If value is None, atrous convolutions will not be used. Default is None.

            pool_sizes (list[int]) : Pooling ratios for PSP pooling. If value is None, PSP pooling will not be used. Default is None.

            encoder_pool (bool) : Whether to include additional PSP pooling layer at end of encoder. Should not be used if last layer has a size of less than 16 pixels. Default is False.
        """
        super().__init__()
        channels = _force_list(channels)
        channels = channels*2 if len(channels) == 1 else channels

        if dilations and len(dilations) != len(hidden): raise ValueError(f"Amount of dilations must equal amount of hidden residual blocks. Given values are {len(dilations)} and {len(hidden)} respectively.")

        if pool_sizes:
            if hidden[0] % len(pool_sizes) != 0: raise ValueError(f"hidden[0] must be divisible by len(pool_sizes). Given values are {hidden[0]} and {len(pool_sizes)} respectively.")
            if encoder_pool and hidden[-1] % len(pool_sizes) != 0: raise ValueError(f"hidden[-1] must be divisible by len(pool_sizes) if encoder_pool is True. Given values are {hidden[-1]} and {len(pool_sizes)} respectively.")
        else:
            if encoder_pool: raise ValueError(f"encoder_pool cannot be True if pool_sizes are not provided.")

        self.norm = nn.BatchNorm2d(channels[0]) if not dilations else None

        self.encoder, self.decoder = nn.ModuleList(), nn.ModuleList()
        layers = [channels[0], *hidden]
        n_layers = len(layers) - 1
        for layer_idx in range(n_layers):
            self.encoder.append(get_resblock(in_channels=layers[layer_idx], out_channels=layers[layer_idx+1], dilations=dilations[layer_idx] if dilations else None, depth=depth))
            if layer_idx + 1 < n_layers:
                self.decoder.append(get_resblock(in_channels=layers[-layer_idx-1]-int(layers[-layer_idx-2]/2), out_channels=layers[-layer_idx-2], dilations=dilations[-layer_idx-1] if dilations else None, depth=depth))

        self.encoder_pool = PSP_Pooling(hidden[-1], pool_sizes) if pool_sizes and encoder_pool else None
        self.reconstruction_pool = PSP_Pooling(hidden[0], pool_sizes) if pool_sizes else None

        self.reconstruction = Reconstruction(channels[0], channels[1], hidden[0], scale)

    def forward(self, x):
        x = x / 128 - 1 # Scale input approx from [0, 255] to [-1, 1]
        if self.norm is not None:
            x = self.norm(x)

        skips = [x]
        for idx, layer in enumerate(self.encoder):
            x = layer(x) # ResBlock

            if idx + 1 < len(self.encoder): # Downscale
                skips.append(x)
                x = F.max_pool2d(x, kernel_size=2)

        if self.encoder_pool is not None:
            x = self.encoder_pool(x)

        for idx, layer in enumerate(self.decoder):
            x = x = F.pixel_shuffle(x, 2) # Upscale

            x = torch.cat([x, skips.pop()], dim=1) # ResBlock
            x = layer(x)

        if self.reconstruction_pool is not None:
            x = self.reconstruction_pool(x)

        x = torch.cat([x, skips.pop()], dim=1) # Final skip connection before reconstruction
        if len(skips) != 0: raise IndexError(f"Skip connection mismatch between encoder and decoder. {len(skips)} skip connections are unused.")

        x = self.reconstruction(x)

        x = x * 128 + 128 # Scale output approx from [-1, 1] to [0, 255]
        return x
    
    def extra_repr(self):
        return f"{'Atrous ' if self.norm is None else ''}ResUNet with {self.reconstruction.scale}x upscaling\n{len(self.encoder)} residual decoder blocks with {self.encoder[0].depth} hidden layers each\nPSP pooling {'enabled' if self.reconstruction_pool else 'disabled'}"

class ResUNetA():
    def __new__(cls,
            channels : int = 1,
            hidden : list[int] = [64, 128, 256, 512, 1024],
            scale : int = 4,
            depth : int = 3,
            dilations : list[list[int]] = [[1,3,15,31],[1,3,15],[1,3],[1],[1]],
            pool_sizes : list[int] = [1, 2, 4, 8],
            encoder_pool : bool = False,
        ):
        r""":class:`ResUNet` wrapper of Atrous Residual UNet as detailed in Diakogiannis et al., 2019.
        Provides alternative default arguments for an atrous network.

        Channel sizes hidden[0] (and hidden[-1] if encoder_pool is True) must be divisible by pool_sizes.

        Args:
            channels (int) : Number of channels in image data. Can also be a list of in channels and out channels respectively.

            hidden (list[int]) : Elementwise list of channels per residual block controlling width and length of model.

            scale (int) : Upscaling factor for predictions. Choose a power of 2 for best results. Default is 4.

            depth (int) : Number of hidden layers per residual block. Default is 3.

            dilations (list[list[int]]) : List of dilation values per layer. If value is None, atrous convolutions will not be used. Default is [[1,3,15,31],[1,3,15],[1,3],[1],[1]].

            pool_sizes (list[int]) : Pooling ratios for PSP pooling. If value is None, PSP pooling will not be used. Default is [1, 2, 4, 8].

            encoder_pool (bool) : Whether to include additional PSP pooling layer at end of encoder. Should not be used if last layer has a size of less than 16 pixels. Default is False.
        """
        return ResUNet(
            channels,
            hidden,
            scale,
            depth,
            dilations,
            pool_sizes,
            encoder_pool,
        )



import torch, os, random
import torch.nn as nn
import torch.nn.functional as F
import numpy as np
from torch.utils.data import DataLoader, Dataset
from pytorch_msssim import ssim
from skopt import gp_minimize
from skopt.space import Dimension
from tqdm import tqdm
from PIL import Image
from .crappifiers import Crappifier
from .data import _RandomIterIdx, _invert_idx
from .util import SSIMLoss, _get_callbacks, pixel_metric, _psnr_metric
from .predict import _collage_preds
from .models._blocks import GradHist

Image.MAX_IMAGE_PIXELS = None

def train_paired(
        model : nn.Module,
        dataset : Dataset,
        batch_size : int,
        loss_fn : nn.Module,
        optim : torch.optim.Optimizer,
        epochs : int,
        device : str = "cpu",
        scheduler : torch.optim.lr_scheduler = None,
        log_frequency : int = 50,
        checkpoint_dir : str = None,
        collage_dir : str = None,
        clamp : bool = False,
        dataloader_kwargs = None,
        callbacks = None,
    ):
    r"""Trains model on paired high-low-resolution crappified data.

    Args:
        model (nn.Module) : Model to train on paired data.

        dataset (Dataset) : Paired image dataset to load data from.

        batch_size (int) : Batch size for dataloader.

        loss_fn (nn.Module) : Loss function for loss calculation.

        optim (Optimizer) : Optimizer for weight calculation.

        epochs (int) : Number of epochs to train model for.

        device (str) : Device to train model on. Default is "cpu".

        scheduler (LRScheduler) : Optional learning rate scheduler for training. Default is None.

        log_frequency (int) : Frequency to log losses and recalculate metrics in steps. Default is 50.

        checkpoint_dir (str) : Directory to save model checkpoints each epoch. A value of None skips checkpointing. Default is None.

        collage_dir (str) : Directory to save validation collages each epoch. A value of None skips the collage. Default is None.

        clamp (bool) : Whether to clamp model image output before weight calculation. Default is False.

        dataloader_kwargs (dict[str, Any]) : Keyword arguments for pytorch ``Dataloader``. Default is None.

        callbacks (list[Callable]) : Callbacks after each training batch. Can optionally specify an argument for locals to be passed. Default is None.

    Returns:
        train_losses (list[float]) : List of losses during training.

        val_losses (list[float]) : Validation losses per epoch.
    """
    dataloader_kwargs = {} if dataloader_kwargs is None else dataloader_kwargs
    callbacks, callback_locals = _get_callbacks(callbacks)
    image_range = 255

    train_dataloader = DataLoader(dataset, batch_size, sampler=_RandomIterIdx(_invert_idx(dataset.val_idx, len(dataset))), **dataloader_kwargs)
    val_dataloader = DataLoader(dataset, batch_size, sampler=_RandomIterIdx(dataset.val_idx, seed=True), **dataloader_kwargs)
    include_metric = True if type(scheduler) == torch.optim.lr_scheduler.ReduceLROnPlateau else False

    model.to(device)

    train_losses, val_losses = [], []
    for epoch in range(epochs):
        # Train
        model.train()
        print(f"Epoch {epoch}:")

        progress = tqdm(train_dataloader)
        for batch_idx, (hr, lr) in enumerate(progress):
            hr, lr = hr.to(device), lr.to(device)

            hr_hat = model(lr)
            if clamp:
                hr_hat = torch.clamp(hr_hat, 0, image_range)

            loss = loss_fn(hr_hat/image_range, hr/image_range)
            loss.backward()
            optim.step()
            optim.zero_grad()

            if batch_idx % log_frequency == 0 or batch_idx == len(progress) - 1:
                train_losses.append(loss.item())

                mse = nn.functional.mse_loss(hr_hat/image_range, hr/image_range)
                progress.set_description(f"pixel[{pixel_metric(mse, image_range):.2f}], psnr[{_psnr_metric(mse):.2f}], ssim[{ssim(hr_hat, hr, data_range=image_range):.3f}]")
            
            if batch_idx == max(len(progress), 2) - 2:
                # Accessible from callbacks via locals
                last_full = [lr.cpu(), hr_hat.cpu(), hr.cpu()]
            
            for idx, callback in enumerate(callbacks):
                if callback_locals[idx]:
                    callback(locals())
                else:
                    callback()

        # Validation
        model.eval()
        print(f"Epoch {epoch} validation...")

        val_loss = []
        progress = tqdm(val_dataloader)

        with torch.no_grad():
            for batch_idx, (hr, lr) in enumerate(progress):
                hr, lr = hr.to(device), lr.to(device)

                hr_hat = model(lr)
                if clamp:
                    hr_hat = torch.clamp(hr_hat, 0, image_range)

                loss = loss_fn(hr_hat/image_range, hr/image_range)
                val_loss.append(loss.item())

                if batch_idx == max(len(progress), 2) - 2:
                    last_full_val = [lr.cpu(), hr_hat.cpu(), hr.cpu()]

        val_loss = sum(val_loss) / len(val_loss)
        val_losses.append(val_loss)
        print(f"Epoch {epoch} validation loss: {val_loss:4f}\n")

        if checkpoint_dir and epoch < epochs - 1:
            os.makedirs(checkpoint_dir, exist_ok=True)
            torch.save(model.state_dict(), f"{checkpoint_dir}/checkpoint{epoch}_{model.__class__.__name__}_{val_loss:.4f}.pth")

        if collage_dir:
            collage = _collage_preds(*last_full_val, crop_res=dataset.crop_res, lr_scale=dataset.lr_scale)
            os.makedirs(collage_dir, exist_ok=True)
            collage.save(f"{collage_dir}/epoch{epoch}_loss{val_loss:.4f}.png")

        if scheduler:
            if include_metric:
                scheduler.step(val_loss)
            else:
                scheduler.step()

    return train_losses, val_losses

def train_crappifier(
        model : nn.Module,
        dataset : Dataset,
        batch_size : int,
        optim : torch.optim.Optimizer,
        epochs : int,
        sigma : int = 5,
        clip : float = 3,
        device : str = "cpu",
        scheduler : torch.optim.lr_scheduler = None,
        log_frequency : int = 50,
        checkpoint_dir : str = None,
        collage_dir : str = None,
        clamp : bool = False,
        dataloader_kwargs = None
    ):
    r"""EXPERIMENTAL, NOT CURRENTLY RECOMMENDED FOR MOST WORKFLOWS!
    
    Trains an :class:`nn.Module` model as a crappifier on high-low-resolution paired data.
    The model must output an image the same size as the input/have a `scale` value of 1.
    This is not necessary if you are using a :class:`Crappifier` instance as your crappifier.

    Args:
        model (nn.Module) : Model to train on paired data.

        dataset (Dataset) : Paired image dataset to load data from.

        batch_size (int) : Batch size for dataloader.

        optim (Optimizer) : Optimizer for weight calculation.

        epochs (int) : Number of epochs to train model for.

        sigma (int) : Precision of noise distribution. Higher values will better approximate noise distribution but can cause larger gradients that are unstable during training. Default is 5.

        clip : (float) : Max gradient for gradient clipping. Use None for no clipping. Default is 3.

        device (str) : Device to train model on. Default is "cpu".

        scheduler (LRScheduler) : Optional learning rate scheduler for training. Default is None.

        log_frequency (int) : Frequency to log losses and recalculate metrics in steps. Default is 50.

        checkpoint_dir (str) : Directory to save model checkpoints each epoch. A value of None skips checkpointing. Default is None.

        collage_dir (str) : Directory to save validation collages each epoch. A value of None skips the collage. Default is None.

        clamp (bool) : Whether to clamp model image output before weight calculation. Default is False.

        dataloader_kwargs (dict[str, Any]) : Keyword arguments for pytorch ``Dataloader``. Default is None.

        callbacks (list[Callable]) : Callbacks after each training batch. Can optionally specify an argument for locals to be passed. Default is None.

    Returns:
        train_losses (list[float]) : List of losses during training.

        val_losses (list[float]) : Validation losses per epoch.
    """
    dataloader_kwargs = {} if dataloader_kwargs is None else dataloader_kwargs
    callbacks, callback_locals = _get_callbacks(callbacks)
    image_range = 255

    train_dataloader = DataLoader(dataset, batch_size, sampler=_RandomIterIdx(_invert_idx(dataset.val_idx, len(dataset))), **dataloader_kwargs)
    val_dataloader = DataLoader(dataset, batch_size, sampler=_RandomIterIdx(dataset.val_idx, seed=True), **dataloader_kwargs)
    include_metric = True if type(scheduler) == torch.optim.lr_scheduler.ReduceLROnPlateau else False

    model.to(device)

    # Leave on cpu?
    hist_fn = GradHist(sigma=sigma)
    ssim_loss = SSIMLoss(ms=False)

    train_losses, val_losses = [], []
    for epoch in range(epochs):
        # Train
        model.train()
        print(f"Epoch {epoch}:")

        progress = tqdm(train_dataloader)
        for batch_idx, (hr, lr) in enumerate(progress):
            scale = int(hr.shape[-1]/lr.shape[-1])
            ds_hr = hr[:, :, ::scale, ::scale]
            ds_hr = ds_hr.to(device)

            lr_hat = model(ds_hr)
            if clamp:
                lr_hat = torch.clamp(lr_hat, 0, image_range)

            loss = _crappifier_loss(lr.to(device), lr_hat, ds_hr, hist_fn, ssim_loss)
            loss.backward()

            if clip is not None and clip > 0:
                nn.utils.clip_grad_value_(model.parameters(), clip)

            optim.step()
            optim.zero_grad()

            if batch_idx % log_frequency == 0 or batch_idx == len(progress) - 1:
                train_losses.append(loss.item())

                progress.set_description(f"loss[{loss.item():.4f}]")

            if batch_idx == max(len(progress), 2) - 2:
                # Accessible from callbacks via locals
                last_full = [lr.cpu(), lr_hat.cpu(), hr.cpu()]

            for idx, callback in enumerate(callbacks):
                if callback_locals[idx]:
                    callback(locals())
                else:
                    callback()

        # Validation
        model.eval()
        print(f"Epoch {epoch} validation...")

        val_loss = []
        progress = tqdm(val_dataloader)

        with torch.no_grad():
            for batch_idx, (hr, lr) in enumerate(progress):
                scale = int(hr.shape[-1]/lr.shape[-1])
                ds_hr = hr[:, :, ::scale, ::scale]
                ds_hr = ds_hr.to(device)

                lr_hat = model(ds_hr)
                if clamp:
                    lr_hat = torch.clamp(lr_hat, 0, image_range)

                loss = _crappifier_loss(lr.to(device), lr_hat, ds_hr, hist_fn, ssim_loss)
                val_loss.append(loss.item())

                if batch_idx == max(len(progress), 2) - 2:
                    last_full_val = [lr.cpu(), lr_hat.cpu(), hr.cpu()]

        val_loss = sum(val_loss) / len(val_loss)
        val_losses.append(val_loss)
        print(f"Epoch {epoch} validation loss: {val_loss:4f}\n")

        if checkpoint_dir and epoch < epochs - 1:
            os.makedirs(checkpoint_dir, exist_ok=True)
            torch.save(model.state_dict(), f"{checkpoint_dir}/checkpoint{epoch}_{model.__class__.__name__}_{val_loss:.4f}.pth")

        if collage_dir:
            collage = _collage_preds(*last_full_val, crop_res=dataset.crop_res, lr_scale=dataset.lr_scale)
            os.makedirs(collage_dir, exist_ok=True)
            collage.save(f"{collage_dir}/epoch{epoch}_loss{val_loss:.4f}.png")

        if scheduler:
            if include_metric:
                scheduler.step(val_loss)
            else:
                scheduler.step()

    return train_losses, val_losses

def approximate_crappifier(crappifier : Crappifier, space : list[Dimension], dataset : Dataset, max_images = None, opt_kwargs = None):
    r"""Approximates :class:`Crappifier` parameters from ground truth paired images. Uses Bayesian optimization because Crappifier functions are not differentiable.

    Args:
        crappifier (Crappifier) : Crappifier whose parameter space will be optimized.

        space (list[Dimension]) : List of parameter spaces for each crappifier parameter.

        dataset (Dataset) : Paired image dataset to load data from.

        max_images (int) : Number of image samples to average computations over for each optimization step. Default is None, using all images in dataset.

        opt_kwargs (dict[str, Any]) : Keyword arguments for skopt ``gp_minimize``. Default is None
    """
    space = [space] if type(space) is not list else space
    n_samples = len(dataset) if max_images is None else min(max_images, len(dataset))
    opt_kwargs = {} if opt_kwargs is None else opt_kwargs

    objective = _Crappifier_Objective(crappifier, dataset, n_samples).sample

    result = gp_minimize(objective, space, **opt_kwargs)

    return result

class _Crappifier_Objective():
    def __init__(self, crappifier : Crappifier, dataset : Dataset, n_samples : int):
        self.crappifier = crappifier
        self.dataset = dataset
        self.n_samples = n_samples

    def sample(self, params):
        sample_idx = list(range(len(self.dataset)))
        random.shuffle(sample_idx)

        metrics = []
        for idx in sample_idx[:self.n_samples]:
            # Grab gound truth high and low resolution images
            hr, lr = self.dataset[idx]
            hr, lr = np.asarray(hr, dtype=np.uint8), np.asarray(lr, dtype=np.uint8)
            
            # Downsampled high resolution image is the baseline for noise profile comparison
            ds_hr = np.stack([np.asarray(Image.fromarray(channel).resize(lr.shape[-2:], Image.Resampling.BILINEAR)) for channel in hr])

            # Generate artificial low resolution image using optimized crappifier parameters
            lr_hat = self.crappifier(*params).crappify(ds_hr)
 
            # Generate distribution of noise values for both crappified and ground truth low resolution images
            # NOTE: Cant use SSIM or MSE on images as Crappifier noise levels will converge to zero because a noiseless downscaled image will be closer to ground truth than one with correct amount of noise
            pred_profile = lr_hat.astype(np.float32) - ds_hr.astype(np.float32)
            target_profile = lr.astype(np.float32) - ds_hr.astype(np.float32)

            bins = np.arange(-256, 256)
            pred_dist, _ = np.histogram(pred_profile.flatten(), bins)
            target_dist, _ = np.histogram(target_profile.flatten(), bins)
            
            # Aggregate errors of both noise distribution and mean noise profile value
            # We are not generating the noise, so spacial significance is low and mean value is used (spacial loss would underapproximate the correct amount of noise)
            dist_error = np.mean((target_dist - pred_dist)**2) / (lr.shape[-1]**2)
            value_error = abs(target_profile.mean() - pred_profile.mean())

            loss = dist_error + value_error
            metrics.append(loss)
        return sum(metrics) / len(metrics)

def _crappifier_loss(lr, lr_hat, ds_hr, hist_fn, ssim_loss):
    # Process outlined in approximate_crappifier
    pred_profile = lr_hat - ds_hr
    target_profile = lr - ds_hr

    pred_dist = hist_fn(pred_profile)
    target_dist = hist_fn(target_profile)

    # We are generating the noise, so a spacial criterion must be present (in a lower order as to optimize purely to the "noiseless" profile)
    dist_error = F.mse_loss(pred_dist, target_dist) / (lr.shape[-1]**2)
    profile_error = ssim_loss(pred_profile, target_profile)
    # value_error = F.l1_loss(pred_profile.view(pred_profile.shape[0], -1).mean(1), target_profile.view(target_profile.shape[0], -1).mean(1))

    loss = dist_error * profile_error
    return loss



"""
RDNet
Copyright (c) 2024-present NAVER Cloud Corp.
Apache-2.0

Source code of RDNet has been modified for use in PSSR.
"""

import torch
import torch.nn as nn
from timm.layers import DropPath, LayerNorm2d, EffectiveSEModule
from timm.models import named_apply
from functools import partial

class RDNet(nn.Module):
    def __init__(
        self,
        in_channels=1,
        n_init_features=128,
        patch_size=2,
        growth_rates=(64, 104, 128, 128, 128, 128, 224),
        ds_blocks=(False, True, True, False, False, False, True),
        block_type=["Block", "Block", "BlockESE", "BlockESE", "BlockESE", "BlockESE", "BlockESE"],
        n_blocks=(3, 3, 3, 3, 3, 3, 3),
        bottleneck_width_ratio=4,
        drop_path_rate=0.0,
        transition_compression_ratio=0.5,
        ls_init_value=1e-6,
    ):
        super().__init__()

        block_type = [block_type] * len(growth_rates) if type(block_type) is str else block_type
        block_type = ["BlockESE" if block else "Block" for block in block_type]
            
        n_blocks = [n_blocks] * len(growth_rates) if type(n_blocks) is int else n_blocks

        if not len(growth_rates) == len(ds_blocks): raise ValueError(f"growth_rates and ds_blocks must have the same length. Given values are {len(growth_rates)} and {len(ds_blocks)} respectively.")
        if not len(growth_rates) == len(block_type): raise ValueError(f"growth_rates and block_type must have the same length. Given values are {len(growth_rates)} and {len(block_type)} respectively.")
        if not len(growth_rates) == len(n_blocks): raise ValueError(f"growth_rates and n_blocks must have the same length. Given values are {len(growth_rates)} and {len(n_blocks)} respectively.")

        # Stem
        self.stem = PatchifyStem(in_channels, n_init_features, patch_size=patch_size)

        # Feature encoder
        self.feature_info = []
        self.num_stages = len(growth_rates)
        curr_stride = 4  # stem_stride
        num_features = n_init_features
        dp_rates = [x.tolist() for x in torch.linspace(0, drop_path_rate, sum(n_blocks)).split(n_blocks)]

        dense_stages = []
        for i in range(self.num_stages):
            dense_stage_layers = []
            if i != 0:
                compressed_num_features = int(num_features * transition_compression_ratio / 8) * 8
                k_size = stride = 1
                if ds_blocks[i]:
                    curr_stride *= 2
                    k_size = stride = 2
                dense_stage_layers.append(LayerNorm2d(num_features))
                dense_stage_layers.append(
                    nn.Conv2d(num_features, compressed_num_features, kernel_size=k_size, stride=stride, padding=0)
                )
                num_features = compressed_num_features

            stage = DenseStage(
                num_block=n_blocks[i],
                num_input_features=num_features,
                growth_rate=growth_rates[i],
                bottleneck_width_ratio=bottleneck_width_ratio,
                drop_path_rates=dp_rates[i],
                ls_init_value=ls_init_value,
                block_type=block_type[i],
            )
            dense_stage_layers.append(stage)
            num_features += n_blocks[i] * growth_rates[i]

            if i + 1 == self.num_stages or (i + 1 != self.num_stages and ds_blocks[i + 1]):
                self.feature_info += [
                    dict(
                        num_chs=num_features,
                        reduction=curr_stride,
                        module=f'dense_stages.{i}',
                        growth_rate=growth_rates[i],
                    )
                ]
            dense_stages.append(nn.Sequential(*dense_stage_layers))
        self.dense_stages = nn.ModuleList(dense_stages)

        # Initialize weights
        named_apply(partial(_init_weights), self)

        self.ds_blocks = ds_blocks

    def forward(self, x):
        x = self.stem(x)

        skips = []
        for idx, layer in enumerate(self.dense_stages):
            if self.ds_blocks[idx]: # Skip connection before each downsample
                skips.append(x)
            x = layer(x)
        
        return *skips, x

class PatchifyStem(nn.Module):
    def __init__(self, num_input_channels, num_init_features, patch_size):
        super().__init__()

        self.stem = nn.Sequential(
            nn.Conv2d(num_input_channels, num_init_features, kernel_size=patch_size, stride=patch_size),
            LayerNorm2d(num_init_features),
        )

    def forward(self, x):
        return self.stem(x)
    
class DenseStage(nn.Sequential):
    def __init__(self, num_block, num_input_features, drop_path_rates, growth_rate, **kwargs):
        super().__init__()
        for i in range(num_block):
            layer = DenseBlock(
                num_input_features=num_input_features,
                growth_rate=growth_rate,
                drop_path_rate=drop_path_rates[i],
                block_idx=i,
                **kwargs,
            )
            num_input_features += growth_rate
            self.add_module(f"dense_block{i}", layer)
        self.num_out_features = num_input_features

    def forward(self, init_feature):
        features = [init_feature]
        for module in self:
            new_feature = module(features)
            features.append(new_feature)
        return torch.cat(features, 1)

class DenseBlock(nn.Module):
    def __init__(
        self,
        num_input_features,
        growth_rate,
        bottleneck_width_ratio,
        drop_path_rate,
        rand_gather_step_prob=0.0,
        block_idx=0,
        block_type="Block",
        ls_init_value=1e-6,
    ):
        super().__init__()
        self.drop_path_rate = drop_path_rate
        self.rand_gather_step_prob = rand_gather_step_prob
        self.block_idx = block_idx
        self.growth_rate = growth_rate

        self.gamma = nn.Parameter(ls_init_value * torch.ones(growth_rate)) if ls_init_value > 0 else None
        growth_rate = int(growth_rate)
        inter_chs = int(num_input_features * bottleneck_width_ratio / 8) * 8
        self.drop_path = DropPath(drop_path_rate)
        self.layers = eval(block_type)(
            in_chs=num_input_features,
            inter_chs=inter_chs,
            out_chs=growth_rate,
        )

    def forward(self, x):
        if type(x) is list:
            x = torch.cat(x, 1)
        x = self.layers(x)

        if self.gamma is not None:
            x = x.mul(self.gamma.reshape(1, -1, 1, 1))
        return x

class Block(nn.Module):
    """D == Dw conv, N == Norm, F == Feed Forward, A == Activation"""
    def __init__(self, in_chs, inter_chs, out_chs):
        super().__init__()
        self.layers = nn.Sequential(
            nn.Conv2d(in_chs, in_chs, groups=in_chs, kernel_size=7, stride=1, padding=3),
            LayerNorm2d(in_chs, eps=1e-6),
            nn.Conv2d(in_chs, inter_chs, kernel_size=1, stride=1, padding=0),
            nn.GELU(),
            nn.Conv2d(inter_chs, out_chs, kernel_size=1, stride=1, padding=0),
        )

    def forward(self, x):
        return self.layers(x)

class BlockESE(nn.Module):
    """D == Dw conv, N == Norm, F == Feed Forward, A == Activation"""
    def __init__(self, in_chs, inter_chs, out_chs):
        super().__init__()
        self.layers = nn.Sequential(
            nn.Conv2d(in_chs, in_chs, groups=in_chs, kernel_size=7, stride=1, padding=3),
            LayerNorm2d(in_chs, eps=1e-6),
            nn.Conv2d(in_chs, inter_chs, kernel_size=1, stride=1, padding=0),
            nn.GELU(),
            nn.Conv2d(inter_chs, out_chs, kernel_size=1, stride=1, padding=0),
            EffectiveSEModule(out_chs),
        )

    def forward(self, x):
        return self.layers(x)

def _init_weights(module, name=None):
    if isinstance(module, nn.Conv2d):
        nn.init.kaiming_normal_(module.weight)
    elif isinstance(module, nn.BatchNorm2d):
        nn.init.constant_(module.weight, 1)
        nn.init.constant_(module.bias, 0)



import torch
import torch.nn as nn
import torch.nn.functional as F
import numpy as np

class Reconstruction(nn.Module):
    def __init__(self, in_channels : int, out_channels : int, hidden : int, scale : int = 4):
        super().__init__()

        self.pre = nn.Conv2d(in_channels=hidden + in_channels, out_channels=scale**2 * hidden, kernel_size=3, padding=1)
        self.conv = nn.Conv2d(in_channels=hidden, out_channels=out_channels, kernel_size=3, padding=1)

        self.scale = scale

    def forward(self, x):
        x = F.relu(self.pre(x), inplace=True)
        x = self.conv(F.pixel_shuffle(x, self.scale))
        return x
    
class ResBlock(nn.Module):
    def __init__(self, in_channels : int, out_channels : int, depth : int, norm : bool = True):
        super().__init__()

        self.conv = nn.Sequential()
        
        n_layers = max(depth, 0) + 1
        for layer_idx in range(n_layers):
            self.conv.append(nn.Conv2d(in_channels if layer_idx == 0 else out_channels, out_channels, kernel_size=3, padding=1))
            
            if norm:
                self.conv.append(nn.BatchNorm2d(out_channels))
            if layer_idx + 1 < n_layers:
                self.conv.append(nn.ReLU(inplace=True))

        self.respass = nn.Conv2d(in_channels, out_channels, kernel_size=1)

        self.depth = depth

    def forward(self, x):
        x = F.relu(self.conv(x) + self.respass(x))
        return x

class ResBlockA(nn.Module):
    def __init__(self, in_channels : int, out_channels : int, dilations : list[int], depth : int, norm : bool = True):
        super().__init__()

        self.dilations = nn.ModuleList()
        for dilation in dilations:
            conv = nn.Sequential()

            n_layers = max(depth, 0) + 1
            for layer_idx in range(n_layers):
                if norm:
                    conv.append(nn.BatchNorm2d(in_channels if layer_idx == 0 else out_channels))
                conv.append(nn.ReLU(inplace=True))

                conv.append(nn.Conv2d(in_channels if layer_idx == 0 else out_channels, out_channels, kernel_size=3, padding="same", dilation=dilation))
            self.dilations.append(conv)

        self.respass = nn.Conv2d(in_channels, out_channels, kernel_size=1)

        self.min_size = max(dilations) * 2 + 1
        self.depth = depth

    def forward(self, x):
        if x.shape[-1] < self.min_size: raise ValueError(f"Tensor size {x.shape} is smaller than than dilation kernel size {self.min_size}.")
        x = F.relu(sum([conv(x) for conv in self.dilations]) + self.respass(x))
        return x

class PSP_Pooling(nn.Module):
    def __init__(self, channels, sizes):
        super().__init__()

        small = channels//len(sizes)
        self.convs = nn.ModuleList([nn.Sequential(nn.Conv2d(small, small, kernel_size=1), nn.BatchNorm2d(small)) for size in sizes])

        self.conv_out = nn.Conv2d(channels, channels, kernel_size=1)
        self.norm_out = nn.BatchNorm2d(channels)

        self.sizes = sizes

    def forward(self, x):
        size = x.shape[-2:]

        # Split x along sizes and apply poolings
        x = torch.chunk(x, chunks=len(self.sizes), dim=1)
        x = [F.interpolate(input=F.max_pool2d(x_chunk, kernel_size=self.sizes[idx]), size=size, mode="bilinear") for idx, x_chunk in enumerate(x)]
        x = [F.relu(self.convs[idx](x_chunk)) for idx, x_chunk in enumerate(x)]
        x = torch.concat(x, dim=1)

        x = F.relu(self.norm_out(self.conv_out(x)))
        return x

class GradHist(nn.Module):
    # https://discuss.pytorch.org/t/differentiable-torch-histc/25865/38
    def __init__(self, bins : int = 512, range : list[int, int] = (-256, 256), sigma : int = 5):
        super().__init__()
        assert range[1] > range[0]
        
        self.delta = float(range[1]-range[0]) / float(bins)
        self.centers = float(range[0]) + self.delta * (torch.arange(bins).float() + 0.5)

        self.sigma = sigma

    def forward(self, x):
        batch, size = x.shape[0], np.prod(x.shape[2:])
        x = x.flatten(start_dim=1)[:, np.newaxis, :] - self.centers[:, np.newaxis].to(device=x.device)
        x = torch.sigmoid(x * self.sigma)
        diff = torch.cat([torch.ones((batch, 1, size), device=x.device), x], dim=1) - torch.cat([x, torch.zeros((batch, 1, size), device=x.device)], dim=1)

        diff = diff.sum(dim=-1)
        return diff[:, :-1]
    
def get_resblock(in_channels : int, out_channels : int, dilations : list[int], depth : int, norm : bool = True):
    if dilations:
        return ResBlockA(in_channels, out_channels, dilations, depth, norm)
    return ResBlock(in_channels, out_channels, depth, norm)



import sys
sys.path.append("..")

import torch, argparse
from torch.nn import MSELoss
from pssr.models import ResUNet, ResUNetA, RDResUNet, RDResUNetA, SwinIR
from pssr.data import ImageDataset, SlidingDataset, PairedImageDataset, PairedSlidingDataset
from pssr.crappifiers import MultiCrappifier, Poisson, AdditiveGaussian, SaltPepper
from pssr.util import SSIMLoss, _tab_string
from pssr.train import train_paired
from pssr.predict import predict_images, test_metrics

IS_NAPARI = False

def _handle_declaration(arg, defaults, req=None):
    req = ", ".join(req) if req is not None else ""

    if arg in defaults:
        expression = arg+f"({req})"
    else:
        expression = arg.split("(")[0]+f"({req}"+"(".join(arg.split("(")[1:])

    return eval(expression)

def parse():
    parser = argparse.ArgumentParser(prog="pssr", description="PSSR CLI for basic usage", formatter_class=argparse.ArgumentDefaultsHelpFormatter)

    parser.add_argument("-t", "--train", action="store_true", help="enable train mode")

    parser.add_argument("-dp", "--data-path", type=str, help="specify dataset path")
    parser.add_argument("-dt", "--data-type", type=str, default="ImageDataset", help="specify dataset type")
    parser.add_argument("-mt", "--model-type", type=str, default="ResUNet", help="specify model type")
    parser.add_argument("-mp", "--model-path", type=str, help="specify model path")

    parser.add_argument("-e", "--epochs", type=int, default=10, help="specify number of training epochs")
    parser.add_argument("-b", "--batch-size", type=int, default=16, help="specify batch size")
    parser.add_argument("-lr", "--lr", type=float, default=1e-3, help="specify learning rate")
    parser.add_argument("-p", "--patience", type=int, default=3, help="specify learning rate decay patience")
    parser.add_argument("-mse", "--mse", action="store_true", help="use MSE loss instead of MS-SSIM loss")

    parser.add_argument("-cp", "--checkpoint", action="store_true", help="save model checkpoints during training")
    parser.add_argument("-sl", "--save-losses", action="store_true", help="save training losses")

    return parser

def main():
    parser = parse()
    if len(sys.argv) == 1:
        parser.print_help(sys.stderr)
        return
    args = parser.parse_args()

    if args.data_path is None:
        print("--data-path(-dp) must be provided")
        return
    elif args.model_path is None and not args.train:
        print("--model-path(-mp) must be provided in predict mode")
        return

    model = _handle_declaration(args.model_type, ["ResUNet", "ResUNetA", "RDResUNet", "RDResUNetA", "SwinIR"])
    dataset = _handle_declaration(args.data_type, ["ImageDataset", "SlidingDataset", "PairedImageDataset", "PairedSlidingDataset"], 
        req=[f"'{item.strip()}'" for item in args.data_path.split(",")] + (["val_split=1"] if not args.train else []))
    
    pssr_head(args.train, model, dataset, None, args.epochs, args.batch_size, args.lr, args.patience, args.mse, args.checkpoint, args.save_losses, args.model_path)
    print("\n")

def pssr_head(train, model, dataset, device, epochs, batch_size, lr, patience, loss_fn, checkpoint, losses, model_path, callbacks = None, stage = None, metrics = None):
    # Shared code with napari plugin
    if stage is not None:
        global IS_NAPARI
        IS_NAPARI = True

        global Status
        from pssr.napari.widgets import Status
        stage.emit(Status.LOADING.value)

        model = eval(model)
        dataset = eval(dataset)

    print(f"\nModel:\n{_tab_string(model.extra_repr())}")
    print(f"\nDataset:\n{_tab_string(str(dataset))}")

    if not IS_NAPARI:
        if torch.cuda.is_available():
            device = "cuda"
            print("\nCUDA enabled device detected, running on GPU.")
        else:
            device = "cpu"
            print("\nCUDA enabled device NOT detected, running on CPU.")
    else:
        if device == "cuda":
            if not torch.cuda.is_available(): raise ValueError("CUDA is specified, but CUDA enabled device is not detected")
            print("\nCUDA enabled, running on GPU.")
        else:
            print("\nCUDA disabled, running on CPU.")

    if model_path:
        if str(model_path) == ".": raise ValueError("Attempted to load model from checkpoint, but path is not provided")
        print(f"Loading {model.__class__.__name__} model from {model_path}")
        model.load_state_dict(torch.load(model_path))
    
    if not IS_NAPARI:
        if train:
            func = lambda : _train_meta(model, dataset, device, epochs, batch_size, lr, patience, loss_fn, checkpoint, losses)
        else:
            func = lambda : _predict_meta(model, dataset, device)
    else:
        if train:
            func = lambda : _train_meta(model, dataset, device, epochs, batch_size, lr, patience, loss_fn, checkpoint, losses, callbacks, stage)
        else:
            func = lambda : _predict_meta(model, dataset, device, callbacks, stage, metrics)
    # Move model to cpu after completion, useful for napari
    _cpu_wrapper(func, model)

def _train_meta(model, dataset, device, epochs, batch_size, lr, patience, loss_fn, checkpoint, losses, callbacks = None, stage = None):
    if not IS_NAPARI:
        loss_fn = MSELoss() if loss_fn else SSIMLoss()
    optim = torch.optim.AdamW(model.parameters(), lr=lr)
    scheduler = torch.optim.lr_scheduler.ReduceLROnPlateau(optim, factor=0.1, patience=patience, threshold=5e-3, verbose=True)
    checkpoint_dir = "checkpoints" if checkpoint else None

    kwargs = dict(
        num_workers = 4,
        pin_memory = True,
    )

    if IS_NAPARI:
        stage.emit(Status.PROGRESS_TRAIN.value)

    print("\nTraining model...")
    train_losses, val_losses = train_paired(
        model=model,
        dataset=dataset,
        batch_size=batch_size,
        loss_fn=loss_fn,
        optim=optim,
        epochs=epochs,
        device=device,
        scheduler=scheduler,
        checkpoint_dir=checkpoint_dir,
        dataloader_kwargs=kwargs,
        callbacks=callbacks,
    )
    print("\nTraining complete!")

    save_path = f"{model.__class__.__name__}_{val_losses[-1]:.4f}.pth"
    torch.save(model.state_dict(), save_path)
    print(f"Saved trained model to {save_path}")

    if losses:
        with open(f"{model.__class__.__name__}_train_losses.txt", "w") as file:
            for loss in train_losses:
                file.write(f"{loss:.6f}\n")
        with open(f"{model.__class__.__name__}_val_losses.txt", "w") as file:
            for loss in val_losses:
                file.write(f"{loss:.6f}\n")

def _predict_meta(model, dataset, device, callbacks = None, stage = None, plotter = None):
    if IS_NAPARI:
        stage.emit(Status.PROGRESS_PREDICT.value)

    print("\nPredicting images from low resolution...")
    predict_images(model, dataset, device, norm=not dataset.is_lr, out_dir="preds", callbacks=callbacks)

    if not dataset.is_lr:
        print("\nCalculating metrics...")
        if not IS_NAPARI:
            metrics = test_metrics(model, dataset, device, callbacks=callbacks)

            print("\nMetrics:")
            for metric in metrics:
                print(f"{metric}: {metrics[metric]}")
        else:
            metrics = test_metrics(model, dataset, device, metrics=["psnr", "ssim"], avg=False, callbacks=callbacks)
            plotter.emit([metrics["psnr"], metrics["ssim"]])

def _cpu_wrapper(func, model):
    try:
        out = func()
    except Exception as error:
        model.to("cpu")
        raise error
    model.to("cpu")
    return out

if __name__ == "__main__":
    main()



"""
SwinIR: Image Restoration Using Swin Transformer, https://arxiv.org/abs/2108.10257
Originally Written by Ze Liu, Modified by Jingyun Liang.

Source code of SwinIR has been modified for use in PSSR.
"""

import math
import torch
import torch.nn as nn
import torch.nn.functional as F
import torch.utils.checkpoint as checkpoint
from timm.layers import DropPath, to_2tuple, trunc_normal_
from ..util import _force_list

class SwinIR(nn.Module):
    def __init__(
            self,
            image_size : int = 128,
            channels : list[int] = 1,
            scale : int = 4,
            embed_dim : int = 96,
            mlp_ratio : int = 2,
            depths : list[int] = [4, 4, 4, 4],
            num_heads : list[int] = [6, 6, 6, 6],
            window_size : int = 8,
            patch_size : int = 1,
            upsampler : str = "pixelshuffle",
            qkv_bias : bool = True,
            qk_scale : float = None,
            drop_rate : float = 0,
            attn_drop_rate : float = 0,
            drop_path_rate : float = 0.1,
            norm_layer : nn.Module = nn.LayerNorm,
            ape : bool = False,
            patch_norm : bool = True,
            use_checkpoint : bool = False,
            resi_connection : str = "1conv",
        ):
        """SwinIR as detailed in Liang et al., 2021.
        
        Args:
            image_size (int) : Input image size. 

            channels (list[int]) : Number of channels in image data. Can also be a list of in channels (low-resolution) and out channels (high-resolution) respectively.

            scale (int) : Upscaling factor for predictions. Choose a power of 2 for best results. Default is 4.

            embed_dim (int) : Number of channels in patch embedding dimension. Default is 96.

            mlp_ratio (int) : Ratio of hidden channels to embedding channels. Default is 2.

            depths (list[int]) : Depths for each Swin Transformer layer. Default is 4 by 4 blocks.

            num_heads (list[int]) : Number of attention heads for each Swin Transformer layer. Default is 6 by 4 blocks.

            window_size (int) : Window size for Swin Transformer layers. Default is 8.

            patch_size (int) : Patch size for embedding and Swin Transformer layers. Default is 1.

            upsampler (str) : Upsamling algorithm for outputs. Options are "pixelshuffle", "pixelshuffledirect", "nearest+conv", and None. Default is "pixelshuffle".

            qkv_bias (bool) : Whether to add learnable bias to query. Default is True.

            qk_scale (float) : Override default qk scale of head_dim ** -0.5 if set. Default is None.

            drop_rate (float) : Dropout rate during training. Default is 0.

            attn_drop_rate (float) : Attention dropout rate during training. Default is 0.

            drop_path_rate (float) : Stochastic depth rate. Default is 0.1.

            norm_layer (nn.Module) : Data normalization layer. Default is :class:`nn.LayerNorm`.

            ape (bool) : Whether to add absolute position embedding to patch embedding. Default is False.

            patch_norm (bool) : Whether to use normalization after patch embedding. Default is False.

            use_checkpoint (bool) : Whether to use checkpointing to save memory. Default is False.

            resi_connection (str) : Which convolutional block to use before residual connection. Options are "1conv" or "3conv". Default is "1conv".
        """
        super().__init__()
        if len(depths) != len(num_heads): raise ValueError(f"Lengths of depths and num_heads must be equal. Given lengths are {len(depths)} and {len(num_heads)}.")

        channels = _force_list(channels)
        channels = channels*2 if len(channels) == 1 else channels
        num_in_ch, num_out_ch = channels[0], channels[1]
        num_feat = 64
        self.img_range = 1
        if channels == 3:
            rgb_mean = (0.4488, 0.4371, 0.4040)
            self.mean = torch.Tensor(rgb_mean).view(1, 3, 1, 1)
        else:
            self.mean = torch.zeros(1, 1, 1, 1)
        self.upscale = scale
        self.upsampler = upsampler
        self.window_size = window_size

        #####################################################################################################
        ################################### 1, shallow feature extraction ###################################
        self.conv_first = nn.Conv2d(num_in_ch, embed_dim, 3, 1, 1)

        #####################################################################################################
        ################################### 2, deep feature extraction ######################################
        self.num_layers = len(depths)
        self.embed_dim = embed_dim
        self.ape = ape
        self.patch_norm = patch_norm
        self.num_features = embed_dim
        self.mlp_ratio = mlp_ratio

        # split image into non-overlapping patches
        self.patch_embed = PatchEmbed(
            img_size=image_size, patch_size=patch_size, in_chans=embed_dim, embed_dim=embed_dim,
            norm_layer=norm_layer if self.patch_norm else None)
        num_patches = self.patch_embed.num_patches
        patches_resolution = self.patch_embed.patches_resolution
        self.patches_resolution = patches_resolution

        # merge non-overlapping patches into image
        self.patch_unembed = PatchUnEmbed(
            img_size=image_size, patch_size=patch_size, in_chans=embed_dim, embed_dim=embed_dim,
            norm_layer=norm_layer if self.patch_norm else None)

        # absolute position embedding
        if self.ape:
            self.absolute_pos_embed = nn.Parameter(torch.zeros(1, num_patches, embed_dim))
            trunc_normal_(self.absolute_pos_embed, std=.02)

        self.pos_drop = nn.Dropout(p=drop_rate)

        # stochastic depth
        dpr = [x.item() for x in torch.linspace(0, drop_path_rate, sum(depths))]  # stochastic depth decay rule

        # build Residual Swin Transformer blocks (RSTB)
        self.layers = nn.ModuleList()
        for i_layer in range(self.num_layers):
            layer = RSTB(dim=embed_dim,
                         input_resolution=(patches_resolution[0],
                                           patches_resolution[1]),
                         depth=depths[i_layer],
                         num_heads=num_heads[i_layer],
                         window_size=window_size,
                         mlp_ratio=self.mlp_ratio,
                         qkv_bias=qkv_bias, qk_scale=qk_scale,
                         drop=drop_rate, attn_drop=attn_drop_rate,
                         drop_path=dpr[sum(depths[:i_layer]):sum(depths[:i_layer + 1])],  # no impact on SR results
                         norm_layer=norm_layer,
                         downsample=None,
                         use_checkpoint=use_checkpoint,
                         img_size=img_size,
                         patch_size=patch_size,
                         resi_connection=resi_connection

                         )
            self.layers.append(layer)
        self.norm = norm_layer(self.num_features)

        # build the last conv layer in deep feature extraction
        if resi_connection == '1conv':
            self.conv_after_body = nn.Conv2d(embed_dim, embed_dim, 3, 1, 1)
        elif resi_connection == '3conv':
            # to save parameters and memory
            self.conv_after_body = nn.Sequential(nn.Conv2d(embed_dim, embed_dim // 4, 3, 1, 1),
                                                 nn.LeakyReLU(negative_slope=0.2, inplace=True),
                                                 nn.Conv2d(embed_dim // 4, embed_dim // 4, 1, 1, 0),
                                                 nn.LeakyReLU(negative_slope=0.2, inplace=True),
                                                 nn.Conv2d(embed_dim // 4, embed_dim, 3, 1, 1))

        #####################################################################################################
        ################################ 3, high quality image reconstruction ################################
        if self.upsampler == 'pixelshuffle':
            # for classical SR
            self.conv_before_upsample = nn.Sequential(nn.Conv2d(embed_dim, num_feat, 3, 1, 1),
                                                      nn.LeakyReLU(inplace=True))
            self.upsample = Upsample(scale, num_feat)
            self.conv_last = nn.Conv2d(num_feat, num_out_ch, 3, 1, 1)
        elif self.upsampler == 'pixelshuffledirect':
            # for lightweight SR (to save parameters)
            self.upsample = UpsampleOneStep(scale, embed_dim, num_out_ch,
                                            (patches_resolution[0], patches_resolution[1]))
        elif self.upsampler == 'nearest+conv':
            # for real-world SR (less artifacts)
            self.conv_before_upsample = nn.Sequential(nn.Conv2d(embed_dim, num_feat, 3, 1, 1),
                                                      nn.LeakyReLU(inplace=True))
            self.conv_up1 = nn.Conv2d(num_feat, num_feat, 3, 1, 1)
            if self.upscale == 4:
                self.conv_up2 = nn.Conv2d(num_feat, num_feat, 3, 1, 1)
            self.conv_hr = nn.Conv2d(num_feat, num_feat, 3, 1, 1)
            self.conv_last = nn.Conv2d(num_feat, num_out_ch, 3, 1, 1)
            self.lrelu = nn.LeakyReLU(negative_slope=0.2, inplace=True)
        else:
            # for image denoising and JPEG compression artifact reduction
            self.conv_last = nn.Conv2d(embed_dim, num_out_ch, 3, 1, 1)

        self.apply(self._init_weights)

    def _init_weights(self, m):
        if isinstance(m, nn.Linear):
            trunc_normal_(m.weight, std=.02)
            if isinstance(m, nn.Linear) and m.bias is not None:
                nn.init.constant_(m.bias, 0)
        elif isinstance(m, nn.LayerNorm):
            nn.init.constant_(m.bias, 0)
            nn.init.constant_(m.weight, 1.0)

    def check_image_size(self, x):
        _, _, h, w = x.size()
        mod_pad_h = (self.window_size - h % self.window_size) % self.window_size
        mod_pad_w = (self.window_size - w % self.window_size) % self.window_size
        x = F.pad(x, (0, mod_pad_w, 0, mod_pad_h), 'reflect')
        return x

    def forward_features(self, x):
        x_size = (x.shape[2], x.shape[3])
        x = self.patch_embed(x)
        if self.ape:
            x = x + self.absolute_pos_embed
        x = self.pos_drop(x)

        for layer in self.layers:
            x = layer(x, x_size)

        x = self.norm(x)  # B L C
        x = self.patch_unembed(x, x_size)

        return x

    def forward(self, x):
        H, W = x.shape[2:]
        x = self.check_image_size(x)
        
        self.mean = self.mean.type_as(x)
        x = (x - self.mean) * self.img_range

        if self.upsampler == 'pixelshuffle':
            # for classical SR
            x = self.conv_first(x)
            x = self.conv_after_body(self.forward_features(x)) + x
            x = self.conv_before_upsample(x)
            x = self.conv_last(self.upsample(x))
        elif self.upsampler == 'pixelshuffledirect':
            # for lightweight SR
            x = self.conv_first(x)
            x = self.conv_after_body(self.forward_features(x)) + x
            x = self.upsample(x)
        elif self.upsampler == 'nearest+conv':
            # for real-world SR
            x = self.conv_first(x)
            x = self.conv_after_body(self.forward_features(x)) + x
            x = self.conv_before_upsample(x)
            x = self.lrelu(self.conv_up1(torch.nn.functional.interpolate(x, scale_factor=2, mode='nearest')))
            if self.upscale == 4:
                x = self.lrelu(self.conv_up2(torch.nn.functional.interpolate(x, scale_factor=2, mode='nearest')))
            x = self.conv_last(self.lrelu(self.conv_hr(x)))
        else:
            # for image denoising and JPEG compression artifact reduction
            x_first = self.conv_first(x)
            res = self.conv_after_body(self.forward_features(x_first)) + x_first
            x = x + self.conv_last(res)

        x = x / self.img_range + self.mean

        return x[:, :, :H*self.upscale, :W*self.upscale]
    
    def extra_repr(self):
        return f"SwinIR with {self.upscale}x upscaling\n{self.num_layers} Swin Transformer blocks with embedding size {self.embed_dim}"

class SwinTransformerBlock(nn.Module):
    r""" Swin Transformer Block.

    Args:
        dim (int): Number of input channels.
        input_resolution (tuple[int]): Input resulotion.
        num_heads (int): Number of attention heads.
        window_size (int): Window size.
        shift_size (int): Shift size for SW-MSA.
        mlp_ratio (float): Ratio of mlp hidden dim to embedding dim.
        qkv_bias (bool, optional): If True, add a learnable bias to query, key, value. Default: True
        qk_scale (float | None, optional): Override default qk scale of head_dim ** -0.5 if set.
        drop (float, optional): Dropout rate. Default: 0.0
        attn_drop (float, optional): Attention dropout rate. Default: 0.0
        drop_path (float, optional): Stochastic depth rate. Default: 0.0
        act_layer (nn.Module, optional): Activation layer. Default: nn.GELU
        norm_layer (nn.Module, optional): Normalization layer.  Default: nn.LayerNorm
    """
    def __init__(self, dim, input_resolution, num_heads, window_size=7, shift_size=0,
            mlp_ratio=4., qkv_bias=True, qk_scale=None, drop=0., attn_drop=0., drop_path=0.,
            act_layer=nn.GELU, norm_layer=nn.LayerNorm
        ):
        super().__init__()
        self.dim = dim
        self.input_resolution = input_resolution
        self.num_heads = num_heads
        self.window_size = window_size
        self.shift_size = shift_size
        self.mlp_ratio = mlp_ratio
        if min(self.input_resolution) <= self.window_size:
            # if window size is larger than input resolution, we don't partition windows
            self.shift_size = 0
            self.window_size = min(self.input_resolution)
        if not 0 <= self.shift_size < self.window_size: raise ValueError(f"shift_size must between 0 and window_size. Given values are {shift_size} and {window_size}.")

        self.norm1 = norm_layer(dim)
        self.attn = WindowAttention(
            dim, window_size=to_2tuple(self.window_size), num_heads=num_heads,
            qkv_bias=qkv_bias, qk_scale=qk_scale, attn_drop=attn_drop, proj_drop=drop)

        self.drop_path = DropPath(drop_path) if drop_path > 0. else nn.Identity()
        self.norm2 = norm_layer(dim)
        mlp_hidden_dim = int(dim * mlp_ratio)
        self.mlp = Mlp(in_features=dim, hidden_features=mlp_hidden_dim, act_layer=act_layer, drop=drop)

        if self.shift_size > 0:
            attn_mask = self.calculate_mask(self.input_resolution)
        else:
            attn_mask = None

        self.register_buffer("attn_mask", attn_mask)

    def calculate_mask(self, x_size):
        # calculate attention mask for SW-MSA
        H, W = x_size
        img_mask = torch.zeros((1, H, W, 1))  # 1 H W 1
        h_slices = (slice(0, -self.window_size),
                    slice(-self.window_size, -self.shift_size),
                    slice(-self.shift_size, None))
        w_slices = (slice(0, -self.window_size),
                    slice(-self.window_size, -self.shift_size),
                    slice(-self.shift_size, None))
        cnt = 0
        for h in h_slices:
            for w in w_slices:
                img_mask[:, h, w, :] = cnt
                cnt += 1

        mask_windows = window_partition(img_mask, self.window_size)  # nW, window_size, window_size, 1
        mask_windows = mask_windows.view(-1, self.window_size * self.window_size)
        attn_mask = mask_windows.unsqueeze(1) - mask_windows.unsqueeze(2)
        attn_mask = attn_mask.masked_fill(attn_mask != 0, float(-100.0)).masked_fill(attn_mask == 0, float(0.0))

        return attn_mask

    def forward(self, x, x_size):
        H, W = x_size
        B, L, C = x.shape
        # assert L == H * W, "input feature has wrong size"

        shortcut = x
        x = self.norm1(x)
        x = x.view(B, H, W, C)

        # cyclic shift
        if self.shift_size > 0:
            shifted_x = torch.roll(x, shifts=(-self.shift_size, -self.shift_size), dims=(1, 2))
        else:
            shifted_x = x

        # partition windows
        x_windows = window_partition(shifted_x, self.window_size)  # nW*B, window_size, window_size, C
        x_windows = x_windows.view(-1, self.window_size * self.window_size, C)  # nW*B, window_size*window_size, C

        # W-MSA/SW-MSA (to be compatible for testing on images whose shapes are the multiple of window size
        if self.input_resolution == x_size:
            attn_windows = self.attn(x_windows, mask=self.attn_mask)  # nW*B, window_size*window_size, C
        else:
            attn_windows = self.attn(x_windows, mask=self.calculate_mask(x_size).to(x.device))

        # merge windows
        attn_windows = attn_windows.view(-1, self.window_size, self.window_size, C)
        shifted_x = window_reverse(attn_windows, self.window_size, H, W)  # B H' W' C

        # reverse cyclic shift
        if self.shift_size > 0:
            x = torch.roll(shifted_x, shifts=(self.shift_size, self.shift_size), dims=(1, 2))
        else:
            x = shifted_x
        x = x.view(B, H * W, C)

        # FFN
        x = shortcut + self.drop_path(x)
        x = x + self.drop_path(self.mlp(self.norm2(x)))

        return x

    def extra_repr(self) -> str:
        return f"dim={self.dim}, input_resolution={self.input_resolution}, num_heads={self.num_heads}, " \
               f"window_size={self.window_size}, shift_size={self.shift_size}, mlp_ratio={self.mlp_ratio}"
    
class RSTB(nn.Module):
    """Residual Swin Transformer Block (RSTB).

    Args:
        dim (int): Number of input channels.
        input_resolution (tuple[int]): Input resolution.
        depth (int): Number of blocks.
        num_heads (int): Number of attention heads.
        window_size (int): Local window size.
        mlp_ratio (float): Ratio of mlp hidden dim to embedding dim.
        qkv_bias (bool, optional): If True, add a learnable bias to query, key, value. Default: True
        qk_scale (float | None, optional): Override default qk scale of head_dim ** -0.5 if set.
        drop (float, optional): Dropout rate. Default: 0.0
        attn_drop (float, optional): Attention dropout rate. Default: 0.0
        drop_path (float | tuple[float], optional): Stochastic depth rate. Default: 0.0
        norm_layer (nn.Module, optional): Normalization layer. Default: nn.LayerNorm
        downsample (nn.Module | None, optional): Downsample layer at the end of the layer. Default: None
        use_checkpoint (bool): Whether to use checkpointing to save memory. Default: False.
        img_size: Input image size.
        patch_size: Patch size.
        resi_connection: The convolutional block before residual connection.
    """
    def __init__(self, dim, input_resolution, depth, num_heads, window_size,
            mlp_ratio=4., qkv_bias=True, qk_scale=None, drop=0., attn_drop=0.,
            drop_path=0., norm_layer=nn.LayerNorm, downsample=None, use_checkpoint=False,
            img_size=224, patch_size=4, resi_connection='1conv'
        ):
        super().__init__()

        self.dim = dim
        self.input_resolution = input_resolution

        self.residual_group = BasicLayer(dim=dim,
                                         input_resolution=input_resolution,
                                         depth=depth,
                                         num_heads=num_heads,
                                         window_size=window_size,
                                         mlp_ratio=mlp_ratio,
                                         qkv_bias=qkv_bias, qk_scale=qk_scale,
                                         drop=drop, attn_drop=attn_drop,
                                         drop_path=drop_path,
                                         norm_layer=norm_layer,
                                         downsample=downsample,
                                         use_checkpoint=use_checkpoint)

        if resi_connection == '1conv':
            self.conv = nn.Conv2d(dim, dim, 3, 1, 1)
        elif resi_connection == '3conv':
            # to save parameters and memory
            self.conv = nn.Sequential(nn.Conv2d(dim, dim // 4, 3, 1, 1), nn.LeakyReLU(negative_slope=0.2, inplace=True),
                                      nn.Conv2d(dim // 4, dim // 4, 1, 1, 0),
                                      nn.LeakyReLU(negative_slope=0.2, inplace=True),
                                      nn.Conv2d(dim // 4, dim, 3, 1, 1))

        self.patch_embed = PatchEmbed(
            img_size=img_size, patch_size=patch_size, in_chans=0, embed_dim=dim,
            norm_layer=None)

        self.patch_unembed = PatchUnEmbed(
            img_size=img_size, patch_size=patch_size, in_chans=0, embed_dim=dim,
            norm_layer=None)

    def forward(self, x, x_size):
        return self.patch_embed(self.conv(self.patch_unembed(self.residual_group(x, x_size), x_size))) + x
    
class BasicLayer(nn.Module):
    """ A basic Swin Transformer layer for one stage.

    Args:
        dim (int): Number of input channels.
        input_resolution (tuple[int]): Input resolution.
        depth (int): Number of blocks.
        num_heads (int): Number of attention heads.
        window_size (int): Local window size.
        mlp_ratio (float): Ratio of mlp hidden dim to embedding dim.
        qkv_bias (bool, optional): If True, add a learnable bias to query, key, value. Default: True
        qk_scale (float | None, optional): Override default qk scale of head_dim ** -0.5 if set.
        drop (float, optional): Dropout rate. Default: 0.0
        attn_drop (float, optional): Attention dropout rate. Default: 0.0
        drop_path (float | tuple[float], optional): Stochastic depth rate. Default: 0.0
        norm_layer (nn.Module, optional): Normalization layer. Default: nn.LayerNorm
        downsample (nn.Module | None, optional): Downsample layer at the end of the layer. Default: None
        use_checkpoint (bool): Whether to use checkpointing to save memory. Default: False.
    """
    def __init__(self, dim, input_resolution, depth, num_heads, window_size,
            mlp_ratio=4., qkv_bias=True, qk_scale=None, drop=0., attn_drop=0.,
            drop_path=0., norm_layer=nn.LayerNorm, downsample=None, use_checkpoint=False
        ):
        super().__init__()
        self.dim = dim
        self.input_resolution = input_resolution
        self.depth = depth
        self.use_checkpoint = use_checkpoint

        # build blocks
        self.blocks = nn.ModuleList([
            SwinTransformerBlock(dim=dim, input_resolution=input_resolution,
                                 num_heads=num_heads, window_size=window_size,
                                 shift_size=0 if (i % 2 == 0) else window_size // 2,
                                 mlp_ratio=mlp_ratio,
                                 qkv_bias=qkv_bias, qk_scale=qk_scale,
                                 drop=drop, attn_drop=attn_drop,
                                 drop_path=drop_path[i] if isinstance(drop_path, list) else drop_path,
                                 norm_layer=norm_layer)
            for i in range(depth)])

        # patch merging layer
        if downsample is not None:
            self.downsample = downsample(input_resolution, dim=dim, norm_layer=norm_layer)
        else:
            self.downsample = None

    def forward(self, x, x_size):
        for blk in self.blocks:
            if self.use_checkpoint:
                x = checkpoint.checkpoint(blk, x, x_size)
            else:
                x = blk(x, x_size)
        if self.downsample is not None:
            x = self.downsample(x)
        return x

    def extra_repr(self) -> str:
        return f"dim={self.dim}, input_resolution={self.input_resolution}, depth={self.depth}"

class WindowAttention(nn.Module):
    r""" Window based multi-head self attention (W-MSA) module with relative position bias.
    It supports both of shifted and non-shifted window.

    Args:
        dim (int): Number of input channels.
        window_size (tuple[int]): The height and width of the window.
        num_heads (int): Number of attention heads.
        qkv_bias (bool, optional):  If True, add a learnable bias to query, key, value. Default: True
        qk_scale (float | None, optional): Override default qk scale of head_dim ** -0.5 if set
        attn_drop (float, optional): Dropout ratio of attention weight. Default: 0.0
        proj_drop (float, optional): Dropout ratio of output. Default: 0.0
    """
    def __init__(self, dim, window_size, num_heads, qkv_bias=True, qk_scale=None, attn_drop=0., proj_drop=0.):
        super().__init__()
        self.dim = dim
        self.window_size = window_size  # Wh, Ww
        self.num_heads = num_heads
        head_dim = dim // num_heads
        self.scale = qk_scale or head_dim ** -0.5

        # define a parameter table of relative position bias
        self.relative_position_bias_table = nn.Parameter(
            torch.zeros((2 * window_size[0] - 1) * (2 * window_size[1] - 1), num_heads))  # 2*Wh-1 * 2*Ww-1, nH

        # get pair-wise relative position index for each token inside the window
        coords_h = torch.arange(self.window_size[0])
        coords_w = torch.arange(self.window_size[1])
        coords = torch.stack(torch.meshgrid([coords_h, coords_w], indexing="ij"))  # 2, Wh, Ww
        coords_flatten = torch.flatten(coords, 1)  # 2, Wh*Ww
        relative_coords = coords_flatten[:, :, None] - coords_flatten[:, None, :]  # 2, Wh*Ww, Wh*Ww
        relative_coords = relative_coords.permute(1, 2, 0).contiguous()  # Wh*Ww, Wh*Ww, 2
        relative_coords[:, :, 0] += self.window_size[0] - 1  # shift to start from 0
        relative_coords[:, :, 1] += self.window_size[1] - 1
        relative_coords[:, :, 0] *= 2 * self.window_size[1] - 1
        relative_position_index = relative_coords.sum(-1)  # Wh*Ww, Wh*Ww
        self.register_buffer("relative_position_index", relative_position_index)

        self.qkv = nn.Linear(dim, dim * 3, bias=qkv_bias)
        self.attn_drop = nn.Dropout(attn_drop)
        self.proj = nn.Linear(dim, dim)

        self.proj_drop = nn.Dropout(proj_drop)

        trunc_normal_(self.relative_position_bias_table, std=.02)
        self.softmax = nn.Softmax(dim=-1)

    def forward(self, x, mask=None):
        """
        Args:
            x: input features with shape of (num_windows*B, N, C)
            mask: (0/-inf) mask with shape of (num_windows, Wh*Ww, Wh*Ww) or None
        """
        B_, N, C = x.shape
        qkv = self.qkv(x).reshape(B_, N, 3, self.num_heads, C // self.num_heads).permute(2, 0, 3, 1, 4)
        q, k, v = qkv[0], qkv[1], qkv[2]  # make torchscript happy (cannot use tensor as tuple)

        q = q * self.scale
        attn = (q @ k.transpose(-2, -1))

        relative_position_bias = self.relative_position_bias_table[self.relative_position_index.view(-1)].view(
            self.window_size[0] * self.window_size[1], self.window_size[0] * self.window_size[1], -1)  # Wh*Ww,Wh*Ww,nH
        relative_position_bias = relative_position_bias.permute(2, 0, 1).contiguous()  # nH, Wh*Ww, Wh*Ww
        attn = attn + relative_position_bias.unsqueeze(0)

        if mask is not None:
            nW = mask.shape[0]
            attn = attn.view(B_ // nW, nW, self.num_heads, N, N) + mask.unsqueeze(1).unsqueeze(0)
            attn = attn.view(-1, self.num_heads, N, N)
            attn = self.softmax(attn)
        else:
            attn = self.softmax(attn)

        attn = self.attn_drop(attn)

        x = (attn @ v).transpose(1, 2).reshape(B_, N, C)
        x = self.proj(x)
        x = self.proj_drop(x)
        return x

    def extra_repr(self) -> str:
        return f'dim={self.dim}, window_size={self.window_size}, num_heads={self.num_heads}'

class PatchMerging(nn.Module):
    r""" Patch Merging Layer.

    Args:
        input_resolution (tuple[int]): Resolution of input feature.
        dim (int): Number of input channels.
        norm_layer (nn.Module, optional): Normalization layer.  Default: nn.LayerNorm
    """

    def __init__(self, input_resolution, dim, norm_layer=nn.LayerNorm):
        super().__init__()
        self.input_resolution = input_resolution
        self.dim = dim
        self.reduction = nn.Linear(4 * dim, 2 * dim, bias=False)
        self.norm = norm_layer(4 * dim)

    def forward(self, x):
        """
        x: B, H*W, C
        """
        H, W = self.input_resolution
        B, L, C = x.shape
        assert L == H * W, "input feature has wrong size"
        assert H % 2 == 0 and W % 2 == 0, f"x size ({H}*{W}) are not even."

        x = x.view(B, H, W, C)

        x0 = x[:, 0::2, 0::2, :]  # B H/2 W/2 C
        x1 = x[:, 1::2, 0::2, :]  # B H/2 W/2 C
        x2 = x[:, 0::2, 1::2, :]  # B H/2 W/2 C
        x3 = x[:, 1::2, 1::2, :]  # B H/2 W/2 C
        x = torch.cat([x0, x1, x2, x3], -1)  # B H/2 W/2 4*C
        x = x.view(B, -1, 4 * C)  # B H/2*W/2 4*C

        x = self.norm(x)
        x = self.reduction(x)

        return x

    def extra_repr(self) -> str:
        return f"input_resolution={self.input_resolution}, dim={self.dim}"

class PatchEmbed(nn.Module):
    r""" Image to Patch Embedding

    Args:
        img_size (int): Image size.  Default: 224.
        patch_size (int): Patch token size. Default: 4.
        in_chans (int): Number of input image channels. Default: 3.
        embed_dim (int): Number of linear projection output channels. Default: 96.
        norm_layer (nn.Module, optional): Normalization layer. Default: None
    """
    def __init__(self, img_size=224, patch_size=4, in_chans=3, embed_dim=96, norm_layer=None):
        super().__init__()
        img_size = to_2tuple(img_size)
        patch_size = to_2tuple(patch_size)
        patches_resolution = [img_size[0] // patch_size[0], img_size[1] // patch_size[1]]
        self.img_size = img_size
        self.patch_size = patch_size
        self.patches_resolution = patches_resolution
        self.num_patches = patches_resolution[0] * patches_resolution[1]

        self.in_chans = in_chans
        self.embed_dim = embed_dim

        if norm_layer is not None:
            self.norm = norm_layer(embed_dim)
        else:
            self.norm = None

    def forward(self, x):
        x = x.flatten(2).transpose(1, 2)  # B Ph*Pw C
        if self.norm is not None:
            x = self.norm(x)
        return x

class PatchUnEmbed(nn.Module):
    r""" Image to Patch Unembedding

    Args:
        img_size (int): Image size.  Default: 224.
        patch_size (int): Patch token size. Default: 4.
        in_chans (int): Number of input image channels. Default: 3.
        embed_dim (int): Number of linear projection output channels. Default: 96.
        norm_layer (nn.Module, optional): Normalization layer. Default: None
    """

    def __init__(self, img_size=224, patch_size=4, in_chans=3, embed_dim=96, norm_layer=None):
        super().__init__()
        img_size = to_2tuple(img_size)
        patch_size = to_2tuple(patch_size)
        patches_resolution = [img_size[0] // patch_size[0], img_size[1] // patch_size[1]]
        self.img_size = img_size
        self.patch_size = patch_size
        self.patches_resolution = patches_resolution
        self.num_patches = patches_resolution[0] * patches_resolution[1]

        self.in_chans = in_chans
        self.embed_dim = embed_dim

    def forward(self, x, x_size):
        B, HW, C = x.shape
        x = x.transpose(1, 2).view(B, self.embed_dim, x_size[0], x_size[1])  # B Ph*Pw C
        return x

class Upsample(nn.Sequential):
    """Upsample module.

    Args:
        scale (int): Scale factor. Supported scales: 2^n and 3.
        num_feat (int): Channel number of intermediate features.
    """

    def __init__(self, scale, num_feat):
        m = []
        if (scale & (scale - 1)) == 0:  # scale = 2^n
            for _ in range(int(math.log(scale, 2))):
                m.append(nn.Conv2d(num_feat, 4 * num_feat, 3, 1, 1))
                m.append(nn.PixelShuffle(2))
        elif scale == 3:
            m.append(nn.Conv2d(num_feat, 9 * num_feat, 3, 1, 1))
            m.append(nn.PixelShuffle(3))
        else:
            raise ValueError(f"scale {scale} is not supported. Supported scales are 2^n and 3.")
        super().__init__(*m)

class UpsampleOneStep(nn.Sequential):
    """UpsampleOneStep module (the difference with Upsample is that it always only has 1conv + 1pixelshuffle)
       Used in lightweight SR to save parameters.

    Args:
        scale (int): Scale factor. Supported scales: 2^n and 3.
        num_feat (int): Channel number of intermediate features.

    """

    def __init__(self, scale, num_feat, num_out_ch, input_resolution=None):
        self.num_feat = num_feat
        self.input_resolution = input_resolution
        m = []
        m.append(nn.Conv2d(num_feat, (scale ** 2) * num_out_ch, 3, 1, 1))
        m.append(nn.PixelShuffle(scale))
        super().__init__(*m)

class Mlp(nn.Module):
    def __init__(self, in_features, hidden_features=None, out_features=None, act_layer=nn.GELU, drop=0.):
        super().__init__()
        out_features = out_features or in_features
        hidden_features = hidden_features or in_features
        self.fc1 = nn.Linear(in_features, hidden_features)
        self.act = act_layer()
        self.fc2 = nn.Linear(hidden_features, out_features)
        self.drop = nn.Dropout(drop)

    def forward(self, x):
        x = self.fc1(x)
        x = self.act(x)
        x = self.drop(x)
        x = self.fc2(x)
        x = self.drop(x)
        return x

def window_partition(x, window_size):
    """
    Args:
        x: (B, H, W, C)
        window_size (int): window size

    Returns:
        windows: (num_windows*B, window_size, window_size, C)
    """
    B, H, W, C = x.shape
    x = x.view(B, H // window_size, window_size, W // window_size, window_size, C)
    windows = x.permute(0, 1, 3, 2, 4, 5).contiguous().view(-1, window_size, window_size, C)
    return windows

def window_reverse(windows, window_size, H, W):
    """
    Args:
        windows: (num_windows*B, window_size, window_size, C)
        window_size (int): Window size
        H (int): Height of image
        W (int): Width of image

    Returns:
        x: (B, H, W, C)
    """
    B = int(windows.shape[0] / (H * W / window_size / window_size))
    x = windows.view(B, H // window_size, W // window_size, window_size, window_size, -1)
    x = x.permute(0, 1, 3, 2, 4, 5).contiguous().view(B, H, W, -1)
    return x



import torch, math, inspect
import torch.nn.functional as F
import torch.nn as nn
import numpy as np
from pytorch_msssim import SSIM, MS_SSIM

class SSIMLoss(nn.Module):
    def __init__(self, channels : int = 1, mix : float = .8, win_size : int = 11, win_sigma : float = 1.5, ms : bool = True, kwargs = None):
        r"""SSIM and MS-SSIM loss with Mix as detailed in Zhao et al., 2018.

        Args:
            channels (int) : Number of channels in image. Default is 1.

            mix (float) : Mix of SSIM loss in loss calculation. 1 is entirely SSIM, 0 is entirely L1 with Gaussian correction. Default is 0.8.

            win_size (int) : Size of Gaussian window. Must be odd. Default is 11.

            win_sigma (float) : Sigma for distribution of Gaussian window. Default is 1.5.

            ms (bool) : Whether to use MS-SSIM over basic SSIM. Default is True.

            kwargs (dict[str, Any]) : Keyword arguments for ``pytorch_msssim``.
        """
        super().__init__()

        kwargs = {} if kwargs is None else kwargs
        self.ssim = MS_SSIM(channel=channels, win_size=win_size, win_sigma=win_sigma, data_range=1, **kwargs) if ms else SSIM(channel=channels, win_size=win_size, win_sigma=win_sigma, data_range=1, **kwargs)

        if mix < 1:
            # Generate Gaussian window for L1 loss
            coords = torch.arange(win_size, dtype=torch.float) - win_size // 2

            gaussian = torch.exp(-(coords ** 2) / (2 * win_sigma ** 2))
            gaussian /= gaussian.sum()

            self.gaussian = torch.outer(gaussian, gaussian)[np.newaxis, np.newaxis, ...]

        self.channels = channels
        self.win_size = win_size
        self.mix = mix

    def forward(self, input, target):
        x = 1 - self.ssim(input, target)
        if self.mix < 1:
            # Combine SSIM with L1 loss with applied Gaussian window for elementwise multiplication against non-reduced L1 loss
            device = "cpu" if input.get_device() == -1 else input.get_device()
            l1 = F.conv2d(F.l1_loss(input, target, reduction="none"), self.gaussian.to(device), groups=self.channels, padding=(self.win_size-1)//2).mean()  # F.conv2d with Gaussian filter
            x = self.mix*x + (1-self.mix)*l1
        return x

def pixel_metric(mse : float, image_range : int = 255):
    r"""Simple metric for calculating average pixel error.
    
    Args:
        mse (float) : Mean squared error between predicted and ground truth images.

        image_range (int) : Value range of image. Default is 255.
    """
    return math.sqrt(mse) * image_range

def _psnr_metric(mse : float):
    return 20 * torch.log10(1 / torch.sqrt(mse))

def _force_list(item):
    if type(item) is not list:
        try:
            return list(item)
        except:
            return [item]
    return item

def _get_callbacks(raw):
    callbacks = [] if raw is None else _force_list(raw)
    callback_locals = [len([arg for arg in inspect.getfullargspec(callback).args if arg != "self"]) == 1 for callback in callbacks]
    return callbacks, callback_locals

def _tab_string(text):
    lines = text.split("\n")
    indented_lines = ["\t" + line for line in lines]
    return "\n".join(indented_lines)



import torch
import torch.nn as nn
import torch.nn.functional as F
from ._blocks import PSP_Pooling, Reconstruction, get_resblock
from ._rdnet import RDNet
from ..util import _force_list

class RDResUNet(nn.Module):
    def __init__(
            self,
            channels : list[int] = 1,
            hidden : list[int] = [1024, 1024, 512, 256],
            scale : int = 4,
            depth : int = 3,
            dilations : list[list[int]] = None,
            pool_sizes : list[int] = None,
            encoder_pool : bool = False,
            rdnet_init : int = 128,
            growth_rates : list[int] = [64, 104, 128, 128, 128, 128, 224],
            ds_blocks : list[bool] = [False, True, True, False, False, False, True],
            ese_blocks : list[bool] = [False, False, True, True, True, True, True],
            n_blocks : list[int] = [3, 3, 3, 3, 3, 3, 3],
            patch_size : int = 2,
            bottleneck : int = 4,
            compression : float = 0.5,
            drop_rate : float = 0,
        ):
        r"""A RDNet (Revitalized DenseNet) encoder and ResUNet decoder with an additional image upscaling block. RDNet is detailed in Kim et al., 2024.
        If ``dilations`` is provided, the decoder is instead a Atrous Residual UNet as detailed in Diakogiannis et al., 2019.

        Skip connections from the RDNet encoder to the ResUNet decoder are created before each downsampling layer.

        Channel sizes hidden[0] (and hidden[-1] if encoder_pool is True) must be divisible by pool_sizes if provided.

        Args:
            channels (list[int]) : Number of channels in image data. Can also be a list of in channels (low-resolution) and out channels (high-resolution) respectively.

            hidden (list[int]) : Elementwise list of hidden layer channels of ResUNet decoder. Each element must have a corresponding skip connection in the RDNet encoder, provided after each downsample block.

            scale (int) : Upscaling factor for predictions. Choose a power of 2 for best results. Default is 4.

            depth (int) : Number of hidden layers per decoder residual block. Default is 3.
            
            dilations (list[list[int]]) : List of dilation values per layer. If value is None, atrous convolutions will not be used. Default is None.

            pool_sizes (list[int]) : Pooling ratios for PSP pooling. If value is None, PSP pooling will not be used. Default is None.

            encoder_pool (bool) : Whether to include additional PSP pooling layer at end of encoder. Should not be used if last layer has a size of less than 16 pixels. Default is False.

            rdnet_init (int) : Number of channels in first RDNet layer. Channels grow from this each layer according to growth parameters. Default is 128.

            growth_rates (list[int]) : Growth rates for each RDNet layer, compounding over each layer.

            ds_blocks (list[bool]) : Specifies which RDNet layers downsample image space. Number of downsampling blocks must be one less than ResUNet hidden layers.

            ese_blocks (list[bool]) : Specifies which RDNet layers are ESE blocks.

            n_blocks (list[int]) : Number of blocks per RDNet layer. Can also be an integer to be held constant across all layers. Default is 3.

            patch_size (int) : Patch size for initial patch embedding. Default is 2.

            bottleneck (int) : Bottleneck width ratio for all blocks. Default is 4.

            compression (float) : Transition compression ratio to offset growth. Default is 0.5.

            drop_rate (float) : Dropout rate for DropPath. Default is 0.
        """
        super().__init__()
        channels = _force_list(channels)
        channels = channels*2 if len(channels) == 1 else channels

        if dilations and len(dilations) != len(hidden): raise ValueError(f"Amount of dilations must equal amount of hidden residual blocks. Given values are {len(dilations)} and {len(hidden)} respectively.")

        if pool_sizes:
            if hidden[0] % len(pool_sizes) != 0: raise ValueError(f"hidden[0] must be divisible by len(pool_sizes). Given values are {hidden[0]} and {len(pool_sizes)} respectively.")
            if encoder_pool and hidden[-1] % len(pool_sizes) != 0: raise ValueError(f"hidden[-1] must be divisible by len(pool_sizes) if encoder_pool is True. Given values are {hidden[-1]} and {len(pool_sizes)} respectively.")
        else:
            if encoder_pool: raise ValueError(f"encoder_pool cannot be True if pool_sizes are not provided.")

        self.norm = nn.BatchNorm2d(channels[0]) if not dilations else None

        if sum(ds_blocks) != len(hidden)-1: raise ValueError(f"Number of downsampling blocks must be one less than ResUNet hidden layers. Given {sum(ds_blocks)} downsampling blocks but {len(hidden)} hidden layers.")

        self.encoder = RDNet(channels[0], rdnet_init, patch_size, growth_rates, ds_blocks, ese_blocks, n_blocks, bottleneck, drop_rate, compression)
        skips = [feature["num_chs"] for feature in self.encoder.feature_info]
        skips.reverse()

        if len(skips) != len(hidden): raise ValueError(f"Each encoder skip connection must have a corresponding decoder hidden layer. There are {len(skips)} skip connections but {len(hidden)} hidden layers.")
        
        self.ratios = [1] + [2]*(len(skips)-1) + [patch_size]

        layers = [0, *hidden]
        self.decoder = nn.ModuleList()
        for layer_idx in range(len(layers)-1):
            self.decoder.append(get_resblock(in_channels=layers[layer_idx]//self.ratios[layer_idx]**2+skips[layer_idx], out_channels=layers[layer_idx+1], dilations=dilations[layer_idx] if dilations else None, depth=depth))

        self.encoder_pool = PSP_Pooling(skips[0], pool_sizes) if pool_sizes and encoder_pool else None
        self.reconstruction_pool = PSP_Pooling(hidden[-1]//self.ratios[-1]**2, pool_sizes) if pool_sizes else None

        self.reconstruction = Reconstruction(channels[0], channels[1], hidden[-1]//self.ratios[-1]**2, scale)

        self.skips = skips

    def forward(self, x):
        x = x / 128 - 1 # Scale input approx from [0, 255] to [-1, 1]
        if self.norm is not None:
            x = self.norm(x)

        skips = [x]
        skips.extend(self.encoder(x))

        if self.encoder_pool is not None:
            skips[-1] = self.encoder_pool(skips[-1])

        for idx, layer in enumerate(self.decoder):
            x = torch.cat([x, skips.pop()], dim=1) if idx != 0 else skips.pop()
            x = layer(x)

            x = F.pixel_shuffle(x, self.ratios[idx+1])
        
        if self.reconstruction_pool is not None:
            x = self.reconstruction_pool(x)
        
        x = torch.cat([x, skips.pop()], dim=1)
        if len(skips) != 0: raise IndexError(f"Skip connection mismatch between encoder and decoder. {len(skips)} skip connections are unused.")

        x = self.reconstruction(x)

        x = x * 128 + 128 # Scale output approx from [-1, 1] to [0, 255]
        return x

    def extra_repr(self):
        return f"{'Atrous ' if self.norm is None else ''}RDResUNet with {self.reconstruction.scale}x upscaling\n{len(self.decoder)} residual blocks with {self.decoder[0].depth} hidden layers each\nSkip connection sizes: {self.skips}\nPSP pooling {'enabled' if self.reconstruction_pool else 'disabled'}"

class RDResUNetA():
    def __new__(cls,
            channels : int = 1,
            hidden : list[int] = [1024, 1024, 512, 256],
            scale : int = 4,
            depth : int = 3,
            dilations : list[list[int]] = [[1],[1],[1,3],[1,3,15]],
            pool_sizes : list[int] = [1, 2, 4, 8],
            encoder_pool : bool = False,
            rdnet_init : int = 128,
            growth_rates : list[int] = [64, 104, 128, 128, 128, 128, 224],
            ds_blocks : list[bool] = [False, True, True, False, False, False, True],
            ese_blocks : list[bool] = [False, False, True, True, True, True, True],
            n_blocks : list[int] = (3, 3, 3, 3, 3, 3, 3),
            patch_size : int = 2,
            bottleneck : int = 4,
            compression : float = 0.5,
            drop_rate : float = 0,
        ):
        r""":class:`RDResUNet` wrapper of Atrous Residual UNet as detailed in Diakogiannis et al., 2019.
        Provides alternative default arguments for an atrous network.

        Skip connections from the RDNet encoder to the ResUNet decoder are created before each downsampling layer.

        Channel sizes hidden[0] (and hidden[-1] if encoder_pool is True) must be divisible by pool_sizes if provided.

        Args:
            channels (int) : Number of channels in image data. Can also be a list of in channels and out channels respectively.

            hidden (list[int]) : Elementwise list of hidden layer channels of ResUNet decoder. Each element must have a corresponding skip connection in the RDNet encoder, provided after each downsample block.

            scale (int) : Upscaling factor for predictions. Choose a power of 2 for best results. Default is 4.

            depth (int) : Number of hidden layers per decoder residual block. Default is 3.
            
            dilations (list[list[int]]) : List of dilation values per layer. If value is None, atrous convolutions will not be used. Default is None.

            pool_sizes (list[int]) : Pooling ratios for PSP pooling. If value is None, PSP pooling will not be used. Default is None.

            encoder_pool (bool) : Whether to include additional PSP pooling layer at end of encoder. Should not be used if last layer has a size of less than 16 pixels. Default is False.

            rdnet_init (int) : Number of channels in first RDNet layer. Channels grow from this each layer according to growth parameters. Default is 128.

            growth_rates (list[int]) : Growth rates for each RDNet layer, compounding over each layer.

            ds_blocks (list[bool]) : Specifies which RDNet layers downsample image space. Number of downsampling blocks must be one less than ResUNet hidden layers.

            ese_blocks (list[bool]) : Specifies which RDNet layers are ESE blocks.

            n_blocks (list[int]) : Number of blocks per RDNet layer. Can also be an integer to be held constant across all layers. Default is 3.

            patch_size (int) : Patch size for initial patch embedding. Default is 2.

            bottleneck (int) : Bottleneck width ratio for all blocks. Default is 4.

            compression (float) : Transition compression ratio to offset growth. Default is 0.5.

            drop_rate (float) : Dropout rate for DropPath. Default is 0.
        """
        return RDResUNet(
            channels,
            hidden,
            scale,
            depth,
            dilations,
            pool_sizes,
            encoder_pool,
            rdnet_init,
            growth_rates,
            ds_blocks,
            ese_blocks,
            n_blocks,
            patch_size,
            bottleneck,
            compression,
            drop_rate,
        )
